###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""The ProductionStatusAgent monitors productions for active requests and takes
care to update their status. Initially this is just to handle simulation
requests.

Allowed production status transitions performed by this agent include:

Idle -> ValidatingInput
Idle -> ValidatingOutput

ValidatedOutput -> Completed

ValidatingInput -> RemovingFiles

RemovedFiles -> Completed

Active -> Idle

Testing -> Idle

In addition this also updates request status from Active to Done.

To do: review usage of production API(s) and re-factor into Production Client

AZ 10.14: merged with a part from RequestTrackingAgent to avoid race conditions
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import time
import os
import sqlite3
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor, wait
from itertools import chain

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities.Time import timeThis
from DIRAC.Interfaces.API.Dirac import Dirac
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations

from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from LHCbDIRAC.Interfaces.API.DiracProduction import DiracProduction
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequestClient import ProductionRequestClient

#############################################################################
# The following is used for StandAlone debugging only (outside Agent)
gStandAlone = False                # work in command line without Agent
# gSimulate = gStandAlone and True  # real clients are replaced with simulation
gSimulate = False
gDoRealUpdate = True         # call status updates
gDoRealTracking = True       # update requests progress

KNOWN_TASK_STATES = [
    'Checking',
    'Completed',
    'Created',
    'Matched',
    'Received',
    'Reserved',
    'Rescheduled',
    'Running',
    'Submitted',
    'Waiting',
]

__RCSID__ = "$Id$"


class ProductionStatusAgent(AgentModule):
  """Usual DIRAC agent."""

  def __init__(self, *args, **kwargs):
    """c'tor.

    :param self: self reference
    :param str agentName: name of agent
    :param str loadName: load name of agent
    :param bool baseAgentName: whatever
    :param dict properties: whatever else
    """
    if not gStandAlone:
      AgentModule.__init__(self, *args, **kwargs)
    else:
      self.log = gLogger

    self.dProd = None
    self.dirac = None
    self.prClient = None
    self.tClient = None

    self.simulationTypes = Operations().getValue(
        'Transformations/ExtendableTransfTypes',
        ['MCSimulation', 'Simulation'],
    )

    self.allKnownStates = (
        'RemovedFiles',
        'RemovingFiles',
        'ValidatedOutput',
        'ValidatingInput',
        'Testing',
        'Active',
        'Idle',
    )

    self.notify = True

    if 'DIRAC' in os.environ:
      self.cacheFile = os.path.join(os.getenv('DIRAC'), 'work/ProductionManagement/cache.db')
    else:
      self.cacheFile = os.path.realpath('cache.db')

    # For processing transformations, it can happen that there are some Unused files
    # with which no tasks can be created. The number of such files can be different depending
    # from the module and distribution between centres.
    # So we declare such transformations isIdle() once there is no jobs, no files in other
    # pending states and the number of Unused files was not changed last cyclesTillIdle times
    self.cyclesTillIdle = 1
    self.filesUnused = {}  # <tID: { 'Number': x, 'NotChanged': n }

    self.prMasters = {}  # [ prID: [<subrequests> ...] ]
    self.prSummary = {}
    self.prProds = {}  # <prID>, map production to known request, from _getProductionRequestsProgress
    self.notPrTrans = defaultdict(list)  # transformation without PR, from _getTransformationsState
    self.toUpdate = []

  #############################################################################
  def initialize(self):
    """Sets default values."""
    # shifter
    self.am_setOption('shifterProxy', 'ProductionManager')

    if not gStandAlone:
      self.notify = eval(self.am_getOption('NotifyProdManager', 'True'))

    # Set the clients
    self.dProd = DiracProduction()
    self.dirac = Dirac()
    if gSimulate:
      raise NotImplementedError()
    self.prClient = ProductionRequestClient()
    self.tClient = TransformationClient()

    return S_OK()

  #############################################################################
  def execute(self):
    """The execution method, track requests progress and implement a part of
    Production SM."""
    updatedT = {}  # updated transformations
    updatedPr = []  # updated production requests (excluding tracking updates)

    # Distinguish between leafs and master requests
    # Masters should not appear in the prodReqSummary and they should have no
    # associated productions.
    self.prMasters = {}  # [ prID: [<subrequests> ...] ]
    self.prSummary = {}
    # { <reqID> :
    #     'type', 'master', 'bkTotal', 'prTotal',  - from _getActiveProductionRequests()
    #     'isDone', 'prods': [ <prodIf> : { 'Used', 'Events' } ] - from __getProductionRequestsProgress
    #     'state' for each production - from _getTransformationsState()
    #     'isIdle', 'isProcIdle' for each 'Active' or 'Idle' production,
    #     'isSimulation' - from _getIdleProductionRequestProductions()
    #     'isFinished' - from _applyProductionRequestsLogic()
    #     'filesTotal' - from _getIdleProductionRequestProductions()
    #     'filesProcessed' - from _getIdleProductionRequestProductions()
    #     'filesUnused' - from _getIdleProductionRequestProductions()
    #     'filesMaxReset' - from _getIdleProductionRequestProductions()
    #     'inputIDs' - from _getExtraInfo()
    #     'hasActiveInput' - from _trackProductionRequests()
    # }
    self.prProds = {}  # <prID>, map production to known request, from _getProductionRequestsProgress

    self.notPrTrans = defaultdict(list)  # transformation without PR, from _getTransformationsState

    self.log.info("******************************")
    self.log.info("Collecting required information")
    self.log.info("******************************")

    result = self._getActiveProductionRequests()
    if not result['OK']:
      self.log.error("Aborting cycle", result["Message"])
      return S_OK()
    if len(self.prProds) < len(self.prSummary):
      self.log.fatal(
          "Aborting cycle, as the number of sub-requests can't be larger than the number of productions")
      return S_OK()

    self._getTransformationsState()
    result = self._getIdleProductionRequestProductions()
    if not result['OK']:
      self.log.error("Aborting cycle", result["Message"])
      return S_OK()

    # That is IMPORTANT to do that after we have the transformation status,
    # since Validation can (really???) update BK, rendering MC incomplete
    result = self._trackProductionRequests()  # also updates PR DB
    if not result['OK']:
      self.log.error("Aborting cycle", result["Message"])
      return S_OK()

    self.log.info("******************************")
    self.log.info("Updating Production Requests and related transformations")
    self.log.info("******************************")

    self._applyProductionRequestsLogic(updatedT, updatedPr)

    self.log.info("******************************")
    self.log.info("Updating Production Request for unrelated transformations (replication, etc.)")
    self.log.info("******************************")

    self._applyOtherTransformationsLogic(updatedT)

    self.log.info("*********")
    self.log.info("Reporting")
    self.log.info("*********")

    if updatedT:
      self.log.info('Transformations updated this cycle:')
      for name, value in updatedT.items():
        self.log.info('Transformations %s: %s => %s' % (name, value['from'], value['to']))

    if updatedPr:
      self.log.info(
          "Production Requests updated to Done status:",
          "%s" % (", ".join(str(i) for i in updatedPr))
      )

    if gDoRealUpdate and not gSimulate:
      self._mailProdManager(updatedT, updatedPr)

    self._cleanFilesUnused()

    if gSimulate:
      self.tClient.animate(self.prClient)

    return S_OK()

  #############################################################################

  @timeThis
  def __getProductionRequestsProgress(self):
    """get known progress for Active requests related productions Failures
    there are critical and can enforce wrong logic."""
    self.log.verbose("Collecting old Production Request Progress...")
    result = self.prClient.getAllProductionProgress()
    if not result['OK']:
      return S_ERROR('Could not retrieve production progress summary: %s' % result['Message'])
    progressSummary = result['Value']  # { <prID> : [ <prodId> : { 'Used', 'Events' } ] }

    for prID, summary in self.prSummary.items():
      # Setting it before updating will give grace period before SM ops
      summary['isDone'] = summary['bkTotal'] >= summary['prTotal']
      summary['prods'] = progressSummary.get(prID, {})
      for tID in summary['prods']:
        self.prProds[tID] = prID
    self.log.verbose("Done with old Production Request Progress")
    return S_OK()

  @timeThis
  def _getActiveProductionRequests(self):
    """get 'Active' requests.

    Failures there are critical and can enforce wrong logic
    Note: this method can be moved to the service
    """
    self.log.info("Collecting active production requests...")
    result = self.prClient.getProductionRequestList(0, '', 'ASC', 0, 0, {'RequestState': 'Active'})
    if not result['OK']:
      return S_ERROR('Could not retrieve active production requests: %s' % result['Message'])
    activeMasters = result['Value']['Rows']
    for pr in activeMasters:
      prID = pr['RequestID']
      if pr['HasSubrequest']:
        self.prMasters[prID] = []
        result = self.prClient.getProductionRequestList(prID, '', 'ASC', 0, 0, {})
        if not result['OK']:
          return S_ERROR('Could not get subrequests for production request %s: %s' % (prID, result['Message']))
        for subPr in result['Value']['Rows']:
          subPrID = subPr['RequestID']
          self.prSummary[subPrID] = {
              'type': pr['RequestType'],
              'master': prID,
              'bkTotal': subPr['bkTotal'],
              'prTotal': subPr['rqTotal'],
          }
          self.prMasters[prID].append(subPrID)
      else:
        self.prSummary[prID] = {
            'type': pr['RequestType'],
            'master': 0,
            'bkTotal': pr['bkTotal'],
            'prTotal': pr['rqTotal'],
        }

    result = self.__getProductionRequestsProgress()
    if not result['OK']:
      return result

    self.log.info('Will work with %s productions from %s Active (sub)requests' %
                  (len(self.prProds), len(self.prSummary)))
    self.log.verbose("Done with collecting Active production requests")
    return S_OK()

  @timeThis
  def __getTransformations(self, status):
    """dev function.

    Get the transformations (print info in the meanwhile)
    """
    res = self.tClient.getTransformationWithStatus(status)
    if not res['OK']:
      self.log.error("Failed to get transformations", "%s: %s" % (status, res['Message']))
      raise RuntimeError("Failed to get %s transformations: %s" % (status, res['Message']))
    if not res['Value']:
      self.log.debug('No transformations in %s status' % status)
      return []
    if len(res['Value']) > 20:
      self.log.verbose("The following number of transformations are in %s status: %u" % (status, len(res['Value'])))
    else:
      valOutStr = ', '.join(str(i) for i in res['Value'])
      self.log.verbose("The following transformations are in %s status: %s" % (status, valOutStr))
    return res['Value']

  @timeThis
  def _getTransformationsState(self):
    """get Transformations state (set 'Other' for not interesting states)
    failures to get something are not critical since there is no reaction on
    'Other' state."""
    self.log.info("Collecting transformations state...")
    try:
      # We put 'Finished' for both
      tListCompleted = self.__getTransformations('Completed')
      tListArchived = self.__getTransformations('Archived')
      tListFinished = tListCompleted + tListArchived
      for tID in tListFinished:
        prID = self.prProds.get(tID, None)
        if prID:
          self.prSummary[prID]['prods'][tID]['state'] = 'Finished'

      for state in self.allKnownStates:
        tList = self.__getTransformations(state)
        for tID in tList:
          prID = self.prProds.get(tID, None)
          if prID:
            self.prSummary[prID]['prods'][tID]['state'] = state
          else:
            self.notPrTrans[state].append(tID)
    except RuntimeError as error:
      self.log.error(error)

    for tID, prID in self.prProds.items():
      if 'state' not in self.prSummary[prID]['prods'][tID]:
        self.prSummary[prID]['prods'][tID]['state'] = 'Other'

    self.log.verbose("Done with collecting transformations states")

  def _getStatusCountersBulk(self, tableName, statusKey, transformationIDs):
    """Get the counter named ``statusKey`` for many transformations

    Wrapper around ``TransformationClient.getCounters`` that returns a nested
    dictionary of:
    ``{transformationID: {state1: X, state2: Y, ..., "TotalCreated": X+Y+...}}``

    :param str tableName: Table in the TransformationDB to use
    :param str statusKey: Name of the status column to consider
    :param list transformationIDs: List of transformation IDs to inspect
    :returns: ``dict`` of ``dict`` of ``int``
    """
    # This is for compatibility with TransformationClient.getTransformationStats
    # and TransformationClient.getTransformationTaskStats
    if tableName == "TransformationTasks":
      totalColumnName = "TotalCreated"
    elif tableName == "TransformationFiles":
      totalColumnName = "Total"
    else:
      raise NotImplementedError(tableName)
    self.log.verbose(
        "Getting counters from for", "%s transformations" % len(transformationIDs)
    )
    res = self.tClient.getCounters(
        tableName,
        ["TransformationID", statusKey],
        {'TransformationID': transformationIDs},
    )
    if not res["OK"]:
        raise RuntimeError(res)
    # Ensure totalColumnName is always included in the output
    statusDict = defaultdict(lambda: defaultdict(int, **{totalColumnName: 0}))
    for attrDict, count in res['Value']:
      tID = attrDict["TransformationID"]
      status = attrDict[statusKey]
      statusDict[tID][status] = count
      statusDict[tID][totalColumnName] += count
    # Loop over transformationIDs to ensure all IDs are included in the output
    return {int(tID): dict(statusDict[int(tID)]) for tID in transformationIDs}

  def _isIdleCache(self, transIDs):
    """Get the cache dictionaries that need to be passed to __isIdle

    This method is used with __isIdle to minimise the number of RPC calls required::

      transformations, taskStatuses, fileStatuses = self._isIdleCache(tIDs)
      for tID in tIDs:
        isIdle, isProcIdle, isSimulation = self.__isIdle(
            tID, transformations[tID], taskStatuses[tID], fileStatuses[tID]
        )

    :param list of str tIDs: The IDs to get cached data for
    :returns: transformations, taskStatuses, fileStatuses
    """
    self.log.verbose("Filling _isIdleCache caches for", "%s transformations" % len(transIDs))
    result = self.tClient.getTransformations(condDict={'TransformationID': transIDs}, limit=1000)
    if not result['OK']:
      self.log.error('Could not get transformations', result['Message'])
      raise RuntimeError('Could not get transformations')
    transformations = {x["TransformationID"]: x for x in result['Value']}
    taskStatuses = self._getStatusCountersBulk("TransformationTasks", "ExternalStatus", transIDs)
    fileStatuses = self._getStatusCountersBulk("TransformationFiles", "Status", transIDs)
    return transformations, taskStatuses, fileStatuses

  def __isIdle(self, tID, tInfo, tStats, filesStats):
    """Checks if a transformation is idle, is procIdle and either the
    transformation is simulation.

    The ``tInfo``, ``tStats``, ``filesStats`` parameters are passed to improve
    performance by avoiding 3 round trips to the transformation service. See
    ``_isIdleCache`` for more details.

    :param str tID: ID of the transformation to check
    :param dict tInfo: Result of ``TransformationClient.getTransformations``
    :param dict tStats: Result of ``self._getStatusCountersBulk`` for the
                        ``ExternalStatus`` column in the ``TransformationTasks`` table
    :param dict filesStats: Result of ``self._getStatusCountersBulk`` for the
                            ``Status`` column in the ``TransformationFiles`` table
    :returns: ``dict`` of ``dict`` of ``int``
    """
    self.log.debug("Checking if transformation is idle", str(tID))
    if tInfo.get('Type', None) in self.simulationTypes:
      isSimulation = True
      # simulation : go to Idle if
      # only failed and done tasks
      # AND number of tasks created in total == number of tasks submitted
      self.log.verbose("Tasks Stats for %d: %s" % (tID, str(tStats)))
      isIdle = (tStats.get('TotalCreated', 0) > 0) and \
          all(tStats.get(status, 0) == 0 for status in KNOWN_TASK_STATES)
      isProcIdle = isIdle
    else:
      isSimulation = False
      # other transformation type : go to Idle if
      # 0 assigned files, unused files number was not changing during the last cyclesTillIdle time
      # AND only failed and done tasks
      self.log.debug("Files stats: %s" % str(filesStats))
      unused = filesStats.get('Unused', 0)
      unusedInherited = filesStats.get('Unused-inherited', 0)
      oldUnused = self.filesUnused.setdefault(tID, {'Number': -1, 'NotChanged': 0})
      if oldUnused['Number'] == unused:
        oldUnused['NotChanged'] += 1
      else:
        oldUnused['NotChanged'] = 0
        oldUnused['Number'] = unused
      assigned = filesStats.get('Assigned', 0)
      isProcIdle = ((assigned == 0) and ((unused == 0) or (oldUnused['NotChanged'] >= self.cyclesTillIdle)))
      if isProcIdle:
        self.log.debug("Tasks Stats: %s" % str(tStats))
        isProcIdle = all(tStats.get(status, 0) == 0 for status in KNOWN_TASK_STATES)
      isIdle = isProcIdle and (unused == 0) and (unusedInherited == 0)
    return (isIdle, isProcIdle, isSimulation)

  def _getIdleProductionRequestProductions(self):
    """evaluate isIdle and isProcIdle status for all productions we need.

    failures are remembered and are taken into account later
    """
    self.log.verbose("Filling caches for _getIdleProductionRequestProductions...")
    transIDs = [
        str(tID) for tID, prID in self.prProds.items()
        if self.prSummary[prID]['prods'][tID]['state'] in ('Active', 'Idle')
    ]
    try:
      transformations, taskStatuses, fileStatuses = self._isIdleCache(transIDs)
    except RuntimeError:
      return S_ERROR("Failed to get _isIdleCache in _getIdleProductionRequestProductions")

    self.log.verbose("Checking idle productions...")
    for tID, prID in self.prProds.items():
      tInfo = self.prSummary[prID]['prods'][tID]
      if tInfo['state'] in ('Active', 'Idle'):
        isIdle, isProcIdle, isSimulation = self.__isIdle(
            tID, transformations[tID], taskStatuses[tID], fileStatuses[tID]
        )
        tInfo['isIdle'] = 'Yes' if isIdle else 'No'
        tInfo['isProcIdle'] = 'Yes' if isProcIdle else 'No'
        tInfo['isSimulation'] = isSimulation
        tInfo['filesTotal'] = fileStatuses[tID]["Total"]
        tInfo['filesProcessed'] = fileStatuses[tID].get("Processed", 0)
        tInfo['filesUnused'] = fileStatuses[tID].get("Unused", 0)
        tInfo['filesMaxReset'] = fileStatuses[tID].get("MaxReset", 0)
      else:
        tInfo['isIdle'] = 'Unknown'
        tInfo['isProcIdle'] = 'Unknown'
        tInfo['isSimulation'] = False
        tInfo['filesTotal'] = None
        tInfo['filesProcessed'] = 0
        tInfo['filesUnused'] = 0
        tInfo['filesMaxReset'] = 0
    self.log.verbose("Checking idle done")
    return S_OK()

  def _trackProductionRequests(self):
    """contact BK for the current number of processed events failures are
    critical."""
    self.log.info("Updating production requests progress...")

    # Using 10 threads, and waiting for the results before continuing
    futureThreads = []
    with ThreadPoolExecutor(10) as threadPool:
      for tID, prID in self.prProds.items():
        futureThreads.append(threadPool.submit(self._getExtraInfo, tID, prID))
      wait(futureThreads)

    # Update the Production request DB with the number of bookkeeping events
    if self.toUpdate:
      if gDoRealTracking:
        result = self.prClient.updateTrackedProductions(self.toUpdate)
      else:
        result = S_OK()
      if not result['OK']:
        self.log.error(
            'Could not send update to the Production Request System',
            result['Message'])  # that is not critical
      else:
        self.log.verbose('The progress of %s Production Requests is updated' % len(self.toUpdate))
    self.log.info("Production requests progress update is finished")

    # Get the status of the Analysis Productions input transformations
    inputTransformIDs = set()
    for summary in self.prSummary.values():
      for tID, tInfo in summary["prods"].items():
        inputTransformIDs = inputTransformIDs.union(
            x for x in tInfo.get("inputIDs", []) if x not in summary["prods"]
        )
    inputTransformStatuses = {}
    if inputTransformIDs:
      retVal = self.tClient.getTransformations(
          condDict={'TransformationID': list(inputTransformIDs)},
          limit=1000,
      )
      if not retVal["OK"]:
        self.log.error("Failed to call getTransformations", retVal["Message"])
        return S_ERROR("Too dangerous to continue")
      inputTransformStatuses = {d["TransformationID"]: d["Status"] for d in retVal["Value"]}
    for prID, summary in self.prSummary.items():
      for tID, tInfo in summary["prods"].items():
        if "inputIDs" not in tInfo:
          continue
        tInfo["hasActiveInput"] = False
        for inputID in tInfo["inputIDs"]:
          if inputID in summary["prods"]:
            inputState = summary["prods"][inputID]["state"]
          else:
            inputState = inputTransformStatuses[inputID]
          if inputState not in ["Archived", "Completed", "Finished"]:
            self.log.info(
                "Marking hasActiveInput=True for",
                "%s (%s) as input %s has status %s" % (tID, prID, inputID, inputState)
            )
            tInfo["hasActiveInput"] = True

    return S_OK()

  def _getExtraInfo(self, tID, prID):
    tInfo = self.prSummary[prID]['prods'][tID]
    # Get the number of bookkeeping events
    self.log.verbose("Getting BK production progress", "Transformation ID = %d" % tID)
    result = BookkeepingClient().getProductionProducedEvents(tID)
    if result['OK']:
      nEvents = result['Value']
      if nEvents and nEvents != tInfo['Events']:
        self.log.verbose("Updating production %d, with BkEvents %d" % (int(tID), int(nEvents)))
        self.toUpdate.append({'ProductionID': tID, 'BkEvents': nEvents})
        tInfo['Events'] = nEvents
    else:
      self.log.error("Progress is not updated", "%s : %s" % (tID, result['Message']))
      return S_ERROR("Too dangerous to continue")
    # Find the transfromation ID for the input to each Analysis Production
    if self.prSummary[prID]["type"] == "AnalysisProduction":
      retVal = TransformationClient().getBookkeepingQuery(tID)
      if not retVal["OK"]:
        self.log.error("Failed to call getBookkeepingQuery", "%s %s" % (tID, retVal["Message"]))
        return S_ERROR("Too dangerous to continue")
      inputBkQuery = retVal["Value"]
      if "ProductionID" in inputBkQuery:
        tInfo["inputIDs"] = [inputBkQuery["ProductionID"]]
      else:
        # FIXME: Just why?!
        if "DataTakingConditions" in inputBkQuery:
          inputBkQuery["ConditionDescription"] = inputBkQuery["DataTakingConditions"]
        retVal = BookkeepingClient().getProductions(inputBkQuery)
        if not retVal["OK"]:
          self.log.error("Failed to call getProductions", "%s %s" % (tID, retVal["Message"]))
          return S_ERROR("Too dangerous to continue")
        tInfo["inputIDs"] = [x[0] for x in retVal["Value"]["Records"]]

  def _cleanFilesUnused(self):
    """remove old transformations from filesUnused."""
    oldIDs = []
    for tID in self.filesUnused:
      if tID in self.prProds:
        continue
      if all(tID not in IDs for _status, IDs in self.notPrTrans.items()):
        oldIDs.append(tID)
    for tID in oldIDs:
      del self.filesUnused[tID]

  def __updateTransformationStatus(self, tID, origStatus, status, updatedT):
    """This method updates the transformation status and logs the changes for
    each iteration of the agent.

    Most importantly this method only allows status transitions based on
    what the original status should be.
    """
    self.log.info('Changing status for transformation %s to %s' % (tID, status))

    if not gDoRealUpdate:
      updatedT[tID] = {'to': status, 'from': origStatus}
      return

    result = self.tClient.setTransformationParameter(tID, 'Status', status)
    if not result['OK']:
      self.log.error("Failed to update status of transformation", "%s from %s to %s" % (tID, origStatus, status))
    else:
      updatedT[tID] = {'to': status, 'from': origStatus}

  def _mailProdManager(self, updatedT, updatedPr):
    """Notify the production manager of the changes as productions should be
    manually extended in some cases."""
    if not updatedT and not updatedPr:
      self.log.verbose('No changes this cycle, mail will not be sent')
      return

    if self.notify:

      with sqlite3.connect(self.cacheFile) as conn:

        try:
          conn.execute('''CREATE TABLE IF NOT EXISTS ProductionStatusAgentCache(
                        production VARCHAR(64) NOT NULL DEFAULT "",
                        from_status VARCHAR(64) NOT NULL DEFAULT "",
                        to_status VARCHAR(64) NOT NULL DEFAULT "",
                        time VARCHAR(64) NOT NULL DEFAULT ""
                       );''')

          conn.execute('''CREATE TABLE IF NOT EXISTS ProductionStatusAgentReqCache(
                        prod_requests VARCHAR(64) NOT NULL DEFAULT "",
                        time VARCHAR(64) NOT NULL DEFAULT ""
                       );''')

        except sqlite3.OperationalError:
          self.log.error("Could not queue mail")

        for tID, val in updatedT.items():
          conn.execute("INSERT INTO ProductionStatusAgentCache (production, from_status, to_status, time)"
                       " VALUES (?, ?, ?, ?)", (tID, val['from'], val['to'], time.asctime())
                       )

        for prod_request in updatedPr:
          conn.execute("INSERT INTO ProductionStatusAgentReqCache (prod_requests, time) VALUES (?, ?)",
                       (prod_request, time.asctime()))

        conn.commit()

        self.log.info('Mail summary queued for sending')

  def __updateProductionRequestStatus(self, prID, status, updatedPr):
    """This method updates the production request status."""
    self.log.info('Marking Production Request %s as %s' % (prID, status))

    if not gDoRealUpdate:
      updatedPr.append(prID)
      return

    reqClient = ProductionRequestClient(useCertificates=False, timeout=120)
    result = reqClient.updateProductionRequest(int(prID), {'RequestState': status})
    if not result['OK']:
      self.log.error(result)
    else:
      updatedPr.append(prID)

  def _applyOtherTransformationsLogic(self, updatedT):
    """animate not Production Requests related transformations failures are not
    critical."""
    self.log.verbose("Updating requests unrelated transformations...")

    self.log.info('Processing %s requests unrelated transformations in "RemovedFiles" state' %
                  len(self.notPrTrans['RemovedFiles']))
    for tID in self.notPrTrans['RemovedFiles']:
      self.__updateTransformationStatus(tID, 'RemovedFiles', 'Completed', updatedT)

    self.log.info('Processing %s requests unrelated transformations in "Active" state' %
                  len(self.notPrTrans['Active']))
    try:
      transformations, taskStatuses, fileStatuses = self._isIdleCache(self.notPrTrans['Active'])
    except RuntimeError:
      self.log.error('Failed to get _isIdleCache for monitoring "Active" requests')
    else:
      for tID in self.notPrTrans['Active']:
        isIdle, _isProcIdle, _isSimulation = self.__isIdle(
            tID, transformations[tID], taskStatuses[tID], fileStatuses[tID]
        )
        if isIdle:
          self.__updateTransformationStatus(tID, 'Active', 'Idle', updatedT)

    self.log.info('Processing %s requests unrelated transformations in "Idle" state' %
                  len(self.notPrTrans['Idle']))
    try:
      transformations, taskStatuses, fileStatuses = self._isIdleCache(self.notPrTrans['Idle'])
    except RuntimeError:
      self.log.error('Failed to get _isIdleCache for monitoring "Idle" requests')
    else:
      for tID in self.notPrTrans['Idle']:
        isIdle, _isProcIdle, _isSimulation = self.__isIdle(
            tID, transformations[tID], taskStatuses[tID], fileStatuses[tID]
        )
        if not isIdle:
          self.__updateTransformationStatus(tID, 'Idle', 'Active', updatedT)

    self.log.verbose('Requests unrelated transformations update is finished')

  def _isReallyDone(self, summary):
    """Evaluate 'isDone' from current update cycle."""
    # Other production types rely on the number of bookkeeping events
    bkTotal = sum(t['Events'] for t in summary['prods'].values() if t['Used'])
    return bkTotal >= summary['prTotal']

  def _producersAreIdle(self, summary):
    """Return True in case all producers (not 'Used') transformations are Idle,
    Finished or not exist."""
    for _tID, tInfo in summary['prods'].items():
      if tInfo['Used']:
        continue
      if tInfo['isIdle'] != 'Yes' and tInfo['state'] != 'Finished':
        return False
    return True

  def _producersAreProcIdle(self, summary):
    """Return True in case all producers (not 'Used') transformations are
    procIdle or finished or not exist."""
    for _tID, tInfo in summary['prods'].items():
      if tInfo['Used']:
        continue
      if tInfo['isProcIdle'] != 'Yes' and tInfo['state'] != 'Finished':
        return False
    return True

  def _processorsAreProcIdle(self, summary):
    """Return True in case all processors ('Used' or not Sim) transformations
    are procIdle or finished or not exist."""
    for _tID, tInfo in summary['prods'].items():
      if not tInfo['Used'] and tInfo['isSimulation']:
        continue
      if tInfo['isProcIdle'] != 'Yes' and tInfo['state'] != 'Finished':
        return False
    return True

  def _mergersAreDone(self, summary):
    """Return True in case all mergers ('Used') transformations are finished or
    not exist."""
    for _tID, tInfo in summary['prods'].items():
      if not tInfo['Used']:
        continue
      if tInfo['state'] != 'Finished':
        return False
    return True

  def _mergersAreProcIdle(self, summary):
    """Return True in case all mergers ('Used') transformations are procIdle or
    finished or not exist."""
    for _tID, tInfo in summary['prods'].items():
      if not tInfo['Used']:
        continue
      if tInfo['isProcIdle'] != 'Yes' and tInfo['state'] != 'Finished':
        return False
    return True

  def _requestedMoreThenProduced(self, tID, summary):
    """Check that this transformation has registered less events than it was
    requested."""
    if summary['prods'][tID]['Events'] < summary['prTotal']:
      self.log.verbose(" Transformation %s has produced less events, asking for extention " % tID)
      return True
    return False

  def _applyProductionRequestsLogic(self, updatedT, updatedPr):
    """ Apply known logic to transformations related to production requests
        NOTE: we make decision based on BK statistic collected in the PREVIOUS cycle (except isReallyDone)
              and transformation status in THIS cycle BEFORE the Logic is started
    """
    self.log.verbose("Production Requests logic...")

    for prID, summary in self.prSummary.items():
      countFinished = 0
      for tID, tInfo in summary['prods'].items():
        if tInfo['state'] == 'Finished':
          countFinished += 1
        elif tInfo['state'] == 'Idle':
          self._handleStateIdle(tID, tInfo, summary, updatedT)
        elif tInfo['state'] == 'RemovedFiles':
          self._handleStateRemovedFiles(tID, tInfo, summary, updatedT)
        elif tInfo['state'] == 'Active':
          self._handleStateActive(tID, tInfo, summary, updatedT)
        elif tInfo['state'] == 'ValidatedOutput':
          self._handleStateValidatedOutput(tID, tInfo, summary, updatedT)
        elif tInfo['state'] == 'ValidatingInput':
          self._handleStateValidatingInput(tID, tInfo, summary, updatedT)
        elif tInfo['state'] == 'Testing':
          self._handleStateTesting(tID, tInfo, summary, updatedT)

      summary['isFinished'] = countFinished == len(summary['prods'])
      if summary['isFinished'] and not summary['master'] and summary['type'] in ['Simulation', 'AnalysisProduction']:
        self.__updateProductionRequestStatus(prID, 'Done', updatedPr)

    for masterID, prList in self.prMasters.items():
      countFinished = sum(1 for prID in prList if self.prSummary[prID]['isFinished'])
      if countFinished == len(prList):
        self.__updateProductionRequestStatus(masterID, 'Done', updatedPr)

    self.log.verbose("Done with Production Requests logic")

  def _handleStateIdle(self, tID, tInfo, summary, updatedT):
    """Used by _applyProductionRequestsLogic"""
    if tInfo['isIdle'] == 'No':
      # 'Idle' && !isIdle() --> 'Active'
      self.__updateTransformationStatus(tID, 'Idle', 'Active', updatedT)
    elif tInfo['isIdle'] != 'Yes':
      return

    if summary['type'] == 'Simulation':
      if not self._isReallyDone(summary):
        # 'Idle' && isIdle() && isDone for MC logic
        if tInfo['Used']:  # for standard sim requests, only the merge will go to ValidatingOutput
          if self._producersAreIdle(summary):
            self.__updateTransformationStatus(tID, 'Idle', 'ValidatingOutput', updatedT)
          # else
          #  it can happened that MC is !isIdle()
        else:  # for standard sim requests, all but the merge will go to ValidatingInput
          if self._mergersAreProcIdle(summary):
            # Note: 'isSimulation' should not be there (it should stay in 'Active')
            self.__updateTransformationStatus(tID, 'Idle', 'ValidatingInput', updatedT)
          # else:
          #   We wait till mergers finish the job

    elif summary['type'] == 'AnalysisProduction':
      if not tInfo["hasActiveInput"]:
        # The number of bookkeeping events is not well defined for Analysis Productions
        allFilesProcessed = tInfo["filesTotal"] and tInfo["filesTotal"] == tInfo["filesProcessed"]
        if allFilesProcessed:
          if tInfo['Used']:
            self.__updateTransformationStatus(tID, 'Idle', 'ValidatingOutput', updatedT)
          else:
            self.__updateTransformationStatus(tID, 'Idle', 'Completed', updatedT)
        elif tInfo["filesTotal"] == tInfo["filesProcessed"] + tInfo["filesMaxReset"]:
          self.log.warn("Transformation has files in MaxReset", "ID=%s count=%s" % (tID, tInfo["filesMaxReset"]))

  def _handleStateRemovedFiles(self, tID, tInfo, summary, updatedT):
    """Used by _applyProductionRequestsLogic"""
    self.__updateTransformationStatus(tID, 'RemovedFiles', 'Completed', updatedT)

  def _handleStateActive(self, tID, tInfo, summary, updatedT):
    """Used by _applyProductionRequestsLogic"""
    if tInfo['isIdle'] == 'Yes':
      if summary['type'] == 'Simulation':
        # 'Active' && isIdle() for MC logic
        if tInfo['Used'] or not tInfo['isSimulation']:
          # The merger will either wait for MC extention (if !isDone)
          # or will start validation once producers are isIdle()
          self.__updateTransformationStatus(tID, 'Active', 'Idle', updatedT)
        # 'Active' && isIdle() && !Used && isSimulation
        elif self._isReallyDone(summary):
          if self._mergersAreProcIdle(summary):
            self.__updateTransformationStatus(tID, 'Active', 'ValidatingInput', updatedT)
        elif self._processorsAreProcIdle(summary) or self._requestedMoreThenProduced(tID, summary):
          # we are not done yet, extend production
          self.__updateTransformationStatus(tID, 'Active', 'Idle', updatedT)
        # else:
        #  we wait till the situation with mergers is clear
      else:
        # for not MC, use reasonable default
        self.__updateTransformationStatus(tID, 'Active', 'Idle', updatedT)
    elif summary["type"] == 'AnalysisProduction' and not tInfo["hasActiveInput"]:
      if tInfo["filesTotal"] == tInfo["filesProcessed"] + tInfo["filesUnused"]:
        self.__updateTransformationStatus(tID, 'Active', 'Flush', updatedT)

  def _handleStateValidatedOutput(self, tID, tInfo, summary, updatedT):
    """Used by _applyProductionRequestsLogic"""
    if summary['type'] == 'Simulation' and summary['isDone'] and tInfo['Used']:
      # for standard sim requests, only the merge
      self.__updateTransformationStatus(tID, 'ValidatedOutput', 'Completed', updatedT)
    elif summary['type'] == 'AnalysisProduction' and tInfo['Used']:
      # isDone only uses the bookkeeping and therefore isn't useful
      self.__updateTransformationStatus(tID, 'ValidatedOutput', 'Completed', updatedT)
    else:
      self.log.warn("Logical bug: transformation %s unexpectedly has 'ValidatedOutput'" % tID)

  def _handleStateValidatingInput(self, tID, tInfo, summary, updatedT):
    """Used by _applyProductionRequestsLogic"""
    if summary['type'] == 'Simulation' and summary['isDone'] and not tInfo['Used']:
      # for standard sim requests, all but the merge
      self.__updateTransformationStatus(tID, 'ValidatingInput', 'RemovingFiles', updatedT)
    else:
      self.log.warn("Logical bug: transformation %s is unexpectedly 'ValidatingInput'" % tID)

  def _handleStateTesting(self, tID, tInfo, summary, updatedT):
    """Used by _applyProductionRequestsLogic"""
    try:
      # TODO This should ideally be moved out of the loop
      transformations, taskStatuses, fileStatuses = self._isIdleCache([str(tID)])
    except RuntimeError:
      self.log.error("Failed to get _isIdleCache for", str(tID))
      return
    isIdle, isProcIdle, isSimulation = self.__isIdle(
        tID, transformations[tID], taskStatuses[tID], fileStatuses[tID]
    )
    self.log.verbose("TransID %d, %s, %s, %s" % (tID, isIdle, isProcIdle, isSimulation))
    if isIdle:
      self.__updateTransformationStatus(tID, 'Testing', 'Idle', updatedT)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod: ValidateRequest

.. module: ValidateRequest

:synopsis: validateRequest operation handler
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id $"


# # imports
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities import DEncode
from DIRAC.RequestManagementSystem.private.OperationHandlerBase import OperationHandlerBase


class WMSSecureOutputData(OperationHandlerBase):
  """.. class:: ValidateRequest.

  Validate operation handler
  """

  def __init__(self, operation=None, csPath=None):
    """c'tor.

    :param self: self reference
    :param Operation operation: Operation instance
    :param str csPath: CS path for this handler
    """
    OperationHandlerBase.__init__(self, operation, csPath)

  def __call__(self):
    """It expects to find the reqID in operation.Arguments."""
    try:
      decode = DEncode.decode(self.operation.Arguments)
      self.log.debug(decode)
      gLogger.debug("Validating output")
    except ValueError as error:
      self.log.exception(error)
      self.operation.Error = str(error)
      self.operation.Status = "Failed"
      return S_ERROR(str(error))

    self.operation.Status = "Done"
    return S_OK()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""TransformationAgent is and LHCb class just for overwriting some of the DIRAC
methods."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.ResourceStatusSystem.Client.ResourceManagementClient import ResourceManagementClient
from DIRAC.TransformationSystem.Agent.TransformationAgent import TransformationAgent as DIRACTransformationAgent

from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

AGENT_NAME = 'Transformation/LHCbTransformationAgent'


class TransformationAgent(DIRACTransformationAgent):
  """Extends base class."""

  def initialize(self):
    """LHCb defaults."""
    DIRACTransformationAgent.initialize(self)

    self.pluginLocation = self.am_getOption('PluginLocation',
                                            'LHCbDIRAC.TransformationSystem.Agent.TransformationPlugin')
    self.workDirectory = self.am_getWorkDirectory()
    self.debug = self.am_getOption('verbosePlugin', False)

    return S_OK()

  def _getClients(self):
    """returns the clients used in the threads."""
    res = DIRACTransformationAgent._getClients(self)

    threadTransformationClient = TransformationClient()
    threadRMClient = ResourceManagementClient()
    threadBkk = BookkeepingClient()

    res.update({'TransformationClient': threadTransformationClient,
                'ResourceManagementClient': threadRMClient,
                'BookkeepingClient': threadBkk})

    return res

  def __generatePluginObject(self, plugin, clients):
    """Generates the plugin object."""
    try:
      plugModule = __import__(self.pluginLocation, globals(), locals(), ['TransformationPlugin'])
    except ImportError as x:
      gLogger.exception("%s.__generatePluginObject: Failed to import 'TransformationPlugin'" % AGENT_NAME, '', x)
      return S_ERROR()
    try:
      oPlugin = getattr(plugModule, 'TransformationPlugin')('%s' % plugin,
                                                            dataManager=clients['DataManager'],
                                                            transClient=clients['TransformationClient'],
                                                            bkClient=clients['BookkeepingClient'],
                                                            rmClient=clients['ResourceManagementClient'],
                                                            transInThread=self.transInThread)
    except Exception as x:  # pylint: disable=broad-except
      gLogger.exception("%s.__generatePluginObject: Failed to create %s()." % (AGENT_NAME, plugin), '', x)
      return S_ERROR()
    oPlugin.workDirectory = self.workDirectory
    oPlugin.pluginCallback = self.pluginCallback
    if self.debug:
      oPlugin.setDebug()
    return S_OK(oPlugin)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Storage Type."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC.AccountingSystem.Client.Types.BaseAccountingType import BaseAccountingType

__RCSID__ = "$Id$"


class Storage(BaseAccountingType):
  """Storage as extension of BaseAccountingType."""

  def __init__(self):

    super(Storage, self).__init__()

    self.definitionKeyFields = [('StorageElement', "VARCHAR(32)"),
                                ('Directory', "VARCHAR(32)")
                                ]
    self.definitionAccountingFields = [('LogicalSize', 'BIGINT UNSIGNED'),
                                       ('LogicalFiles', 'BIGINT UNSIGNED'),
                                       ('PhysicalSize', 'BIGINT UNSIGNED'),
                                       ('PhysicalFiles', 'BIGINT UNSIGNED')
                                       ]
    self.bucketsLength = [(86400 * 30 * 6, 86400),  # <6m = 1d
                          (31104000, 604800),  # >6m = 1w
                          ]
    self.checkType()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Extension of DIRAC Task Manager."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC.TransformationSystem.Client.TaskManager import WorkflowTasks

__RCSID__ = "$Id$"

COMPONENT_NAME = 'LHCbTaskManager'


class LHCbWorkflowTasks(WorkflowTasks):
  """A simple LHCb extension to the task manager, for now only used to set the
  runNumber and runMetadata."""

  def _handleInputs(self, oJob, paramsDict):
    """set job inputs (+ metadata)"""
    try:
      if paramsDict['InputData']:
        self.log.verbose('Setting input data to %s' % paramsDict['InputData'])
        self.log.verbose('Setting run number to %s' % str(paramsDict.get('RunNumber')))
        oJob.setInputData(paramsDict['InputData'], runNumber=paramsDict.get('RunNumber'))

        try:
          runMetadata = paramsDict['RunMetadata']
          self.log.verbose('Setting run metadata information to %s' % str(runMetadata))
          oJob.setRunMetadata(runMetadata)
        except KeyError:
          pass

    except KeyError:
      self.log.exception('Could not find input data or a run number')
      raise KeyError('Could not found an input data or a run number')

  #############################################################################

  def _handleRest(self, oJob, paramsDict):
    """add as JDL parameters all the other parameters that are not for inputs
    or destination."""

    for paramName, paramValue in paramsDict.items():
      if paramName not in ('InputData', 'RunNumber', 'RunMetadata', 'Site', 'TargetSE'):
        if paramValue:
          self.log.verbose('Setting %s to %s' % (paramName, paramValue))
          oJob._addJDLParameter(paramName, paramValue)

  #############################################################################



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test class for BookkeepingReport."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=missing-docstring, invalid-name, protected-access

__RCSID__ = "$Id$"

import importlib
from itertools import product

import pytest
from mock import MagicMock

# mocks
from DIRAC.DataManagementSystem.Client.test.mock_DM import dm_mock
from LHCbDIRAC.Workflow.Modules.mock_Commons import prod_id, prod_job_id, wms_job_id, \
    workflowStatus, stepStatus, step_id, step_number,\
    step_commons, wf_commons
from LHCbDIRAC.BookkeepingSystem.Client.test.mock_BookkeepingClient import bkc_mock

# sut
from LHCbDIRAC.Workflow.Modules.BookkeepingReport import BookkeepingReport


rv_m = importlib.import_module("LHCbDIRAC.Workflow.Modules.ModuleBase")
rv_m.RequestValidator = MagicMock()

mock_pxc = MagicMock()
mock_pxc.return_value = dict()
p_m = importlib.import_module("LHCbDIRAC.Resources.Catalog.PoolXMLFile")
p_m.getOutputType = mock_pxc

bkr = BookkeepingReport(bkClient=bkc_mock, dm=dm_mock)

allCombinations = list(product(wf_commons, step_commons))


@pytest.mark.parametrize("wf_cs, s_cs", allCombinations)
def test_execute(wf_cs, s_cs):
  assert bkr.execute(prod_id, prod_job_id, wms_job_id,
                     workflowStatus, stepStatus,
                     wf_cs, s_cs,
                     step_number, step_id, False)['OK'] is True


@pytest.mark.parametrize("output, outputType, xf_o_r, expected", [
    ('/this/is/an/output.txt', 'txt', {'/this/is/an/output.txt': {}}, ('{}', '/this/is/an/output.txt')),
    ('/this/is/an/output.hist', 'hist', {'/this/is/an/output.txt': {}}, ('Unknown', '/this/is/an/output.hist')),
    ('/this/is/an/output.txt', 'txt', {'/this/is/an/OutPut.txt': {}}, ('{}', '/this/is/an/OutPut.txt')),
])
def test__getFileStatsFromXMLSummary(output, outputType, xf_o_r, expected):
  mock_xfo = MagicMock()
  mock_xfo.outputsEvents = xf_o_r
  bkr.xf_o = mock_xfo

  res = bkr._getFileStatsFromXMLSummary(output, outputType)
  assert res == expected



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Script for scanning the popularity table."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from DIRAC.Core.Base import Script

  since = 30
  getAllDatasets = False
  Script.registerSwitch('', 'Since=', '   Number of days to look for (default: %d)' % since)
  Script.registerSwitch('', 'All', '   If used, gets all existing datasets, not only the used ones')
  Script.setUsageMessage('\n'.join([__doc__,
                                    'Usage:',
                                    '  %s [option|cfgfile]' % Script.scriptName, ]))
  Script.parseCommandLine(ignoreErrors=True)
  for switch in Script.getUnprocessedSwitches():
    if switch[0] == 'Since':
      try:
        since = int(switch[1])
      except ValueError:
        pass
    elif switch[0] == 'All':
      getAllDatasets = True

  from LHCbDIRAC.DataManagementSystem.Client.ScanPopularity import scanPopularity
  scanPopularity(since, getAllDatasets)


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test XMLErr.py."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest
import ast
import xml.etree.ElementTree as ET

import LHCbDIRAC.Core.Utilities.XMLErr as XMLErr


class XMLErrTestCase(unittest.TestCase):

  def __init__(self, *args, **kwargs):
    super(XMLErrTestCase, self).__init__(*args, **kwargs)

  def setUp(self):
    self.jobID = 1
    self.prodID = 2
    self.wmsID = 3

  def tearDown(self):
    pass


class TestXMLErr(XMLErrTestCase):

  #########################################
  # Test extractData()

  def test_extractDataWithNoCounter(self):
    rootNoCounter = ET.fromstring("<counters></counters>")
    expected = dict({"Counters": {"ID": {"JobID": 1, "ProductionID": 2, "wmsID": 3}}})
    result = XMLErr.extractData(rootNoCounter, self.jobID, self.prodID, self.wmsID)

    self.assertTrue(result['OK'])
    self.assertEqual(ast.literal_eval(result['Value']), expected)

  def test_extractDataWithStringValue(self):
    rootStringValue = ET.fromstring('<counters><counter name = "test">"string"</counter></counters>')
    result = XMLErr.extractData(rootStringValue, self.jobID, self.prodID, self.wmsID)

    self.assertFalse(result['OK'])

  def test_extractDataWithFloatValue(self):
    rootFloatValues = ET.fromstring('<counters><counter name = "test">1.5</counter></counters>')
    result = XMLErr.extractData(rootFloatValues, self.jobID, self.prodID, self.wmsID)

    self.assertFalse(result['OK'])

  def test_extractDataNormal(self):
    root = ET.fromstring('<counters><counter name = "test1">1</counter><counter name = "test2">-1</counter></counters>')
    expected = dict({"Counters": {"test1": 1, "test2": -1, "ID": {"JobID": 1, "ProductionID": 2, "wmsID": 3}}})
    result = XMLErr.extractData(root, self.jobID, self.prodID, self.wmsID)

    self.assertTrue(result['OK'])
    self.assertEqual(ast.literal_eval(result['Value']), expected)


if __name__ == '__main__':
  suite = unittest.defaultTestLoader.loadTestsFromTestCase(XMLErrTestCase)
  suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestXMLErr))
  testResult = unittest.TextTestRunner(verbosity=2).run(suite)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Prepare a file (data.py) which is consumed by Ganga, containing the input
files as resolved in the workflow."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR, gLogger

from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from LHCbDIRAC.Core.Utilities.GangaDataFile import GangaDataFile


class CreateDataFile(ModuleBase):
  """CreateDataFile class."""

  def __init__(self, bkClient=None, dm=None):
    """simple init."""

    self.log = gLogger.getSubLogger("CreateDataFile")

    super(CreateDataFile, self).__init__(self.log, bkClientIn=bkClient, dm=dm)

    self.gangaFileName = 'data.py'
    self.poolXMLCatName = 'pool_xml_catalog.xml'
    self.persistency = ''
    self.version = __RCSID__

  def _resolveInputVariables(self):
    """By convention the module parameters are resolved here."""

    super(CreateDataFile, self)._resolveInputVariables()
    super(CreateDataFile, self)._resolveInputStep()

  def execute(self, production_id=None, prod_job_id=None, wms_job_id=None,
              workflowStatus=None, stepStatus=None,
              wf_commons=None, step_commons=None,
              step_number=None, step_id=None,
              gdf=None):
    """Just calls GangaDataFile with some parameter."""

    try:

      super(CreateDataFile, self).execute(self.version, production_id, prod_job_id, wms_job_id,
                                          workflowStatus, stepStatus,
                                          wf_commons, step_commons, step_number, step_id)

      if not self._checkWFAndStepStatus():
        return S_OK()

      self._resolveInputVariables()

      if not gdf:
        gdf = GangaDataFile(fileName=self.gangaFileName,
                            xmlcatalog_file=self.poolXMLCatName,
                            log=self.log)

      if not self.inputDataList:
        return S_OK('No data file generated, because no inputs set')

      gdf.generateDataFile(self.inputDataList, persistency=self.persistency)

      return S_OK()

    except Exception as e:  # pylint:disable=broad-except
      self.log.exception("Failure in CreateDataFile execute module", lException=e)
      return S_ERROR(str(e))

    finally:
      super(CreateDataFile, self).finalize(self.version)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Moving toward a templates-less system."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from six.moves import range

import ast

from DIRAC.Core.Base import Script
Script.parseCommandLine()

from DIRAC import gConfig, gLogger, exit as DIRACexit
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequest import ProductionRequest

__RCSID__ = "$Id$"


def fillVisList(vdict, num):

  # Assuming that, if there's only one element in the list of output visibility flags, every step will catch that flag
  if len(vdict) == 1:
    val = vdict[list(vdict)[0]]
    vdict = dict([(str(i), val) for i in range(int(list(vdict)[0]), int(list(vdict)[0]) + num)])
  # Another assumption: if the number of steps is bigger than that of vis flags,
  # then extend the list with the last flag available
  # to fill the "holes"
  # if len(vlist) < len(slist):
  #  vlist.extend( vlist[-1] * (len(slist) - len(vlist)) )

  return vdict


gLogger = gLogger.getSubLogger('LaunchingRequest_run.py')
currentSetup = gConfig.getValue('DIRAC/Setup')

pr = ProductionRequest()

stepsList = ['{{p1Step}}']
stepsList.append('{{p2Step}}')
stepsList.append('{{p3Step}}')
stepsList.append('{{p4Step}}')
stepsList.append('{{p5Step}}')
stepsList.append('{{p6Step}}')
stepsList.append('{{p7Step}}')
stepsList.append('{{p8Step}}')
stepsList.append('{{p9Step}}')
stepsList.append('{{p10Step}}')
stepsList.append('{{p11Step}}')
stepsList.append('{{p12Step}}')
stepsList.append('{{p13Step}}')
stepsList.append('{{p14Step}}')
stepsList.append('{{p15Step}}')
stepsList.append('{{p16Step}}')
stepsList.append('{{p17Step}}')
stepsList.append('{{p18Step}}')
stepsList.append('{{p19Step}}')
stepsList.append('{{p20Step}}')
pr.stepsList = stepsList

###########################################
# Configurable and fixed parameters
###########################################

pr.appendName = '{{WorkflowAppendName#GENERAL: Workflow string to append to production name#1}}'

w = '{{w#----->WORKFLOW: choose what to do#}}'
pr.prodsTypeList = '{{r#---List here which prods you want to create#DataSwimming,DataStripping,Merge}}'.split(',')
p = '{{p#--now map steps to productions#}}'
p1 = '{{p1#-Production 1 steps (e.g. 1,2,3 = first, second, third)#1,2,3}}'
if p1:
  p1 = p1.split(',')
  p1 = [int(x) for x in p1]
  pr.stepsInProds.append(p1)
p2 = '{{p2#-Production 2 steps (e.g. 4,5)#4,5}}'
if p2:
  p2 = p2.split(',')
  p2 = [int(x) for x in p2]
  pr.stepsInProds.append(p2)
else:
  pr.stepsInProds.append([])
p3 = '{{p3#-Production 3 steps (e.g. 6)#}}'
if p3:
  p3 = p3.split(',')
  p3 = [int(x) for x in p3]
  pr.stepsInProds.append(p3)
else:
  pr.stepsInProds.append([])

localTestFlag = ast.literal_eval('{{localTestFlag#GENERAL: Set True for local test#False}}')
validationFlag = ast.literal_eval('{{validationFlag#GENERAL: Set True for validation prod#False}}')

# workflow params for all productions
pr.startRun = int('{{startRun#GENERAL: run start, to set the start run#0}}')
pr.endRun = int('{{endRun#GENERAL: run end, to set the end of the range#0}}')
pr.runsList = '{{runsList#GENERAL: discrete list of run numbers (do not mix with start/endrun)#}}'
pr.targets = ['{{WorkflowDestination#GENERAL: Workflow destination site e.g. LCG.CERN.cern#}}'] * len(pr.prodsTypeList)
extraOptions = '{{extraOptions#GENERAL: extra options as python dict stepNumber:options#}}'
if extraOptions:
  pr.extraOptions = ast.literal_eval(extraOptions)
pr.derivedProduction = int('{{AncestorProd#GENERAL: ancestor prod to be derived#0}}')
pr.previousProdID = int('{{previousProdID#GENERAL: previous prod ID (for BK query)#0}}')
modulesList = '{{modulesList#GENERAL: custom modules list#}}'
enablePopularityReport = ast.literal_eval('{{popularityReport#GENERAL: enable popularity report#False}}')
pr.visibility = '{{visibilityFlag#GENERAL: visibility flag for input BK Query#Yes}}'

# p1 params
p1Plugin = '{{p1PluginType#PROD-P1: production plugin name#LHCbStandard}}'
p1Priority = int('{{p1Priority#PROD-P1: priority#2}}')
p1CPU = '{{p1MaxCPUTime#PROD-P1: Max CPU time in secs#1000000}}'
p1GroupSize = '{{p1GroupSize#PROD-P1: Group size or number of files per job#1}}'
p1DataSE = '{{p1DataSE#PROD-P1: Output Data Storage Element#Tier1-DST}}'
try:
  p1DataSESpecial = ast.literal_eval('{{p1DataSESpecial#PROD-P1: Special SEs per file type, e.g. {"T1":"SE1"}#}}')
except SyntaxError:
  p1DataSESpecial = {}
p1Policy = '{{p1Policy#PROD-P1: data policy (download or protocol)#download}}'
p1RemoveInputs = ast.literal_eval('{{p1RemoveInputs#PROD-P1: removeInputs flag#False}}')
p1StepMask = '{{p1StepMask#PROD-P1: step output to save, semicolon separated (default is last)#}}'
p1FileMask = '{{p1FileMask#PROD-P1: file types in output to save, semicolon separated (default is all)#}}'
p1NoBKQuery = ast.literal_eval('{{P1NoBKQuery#PROD-P1: run without an input Bookkeeping Query#False}}')
p1multicoreFlag = '{{p1MulticoreFLag#PROD-P1: multicore flag#True}}'
p1NumberOfProcessors = '{{p1NumberOfProcessors#PROD-P1: jobs min/max n of processors#0,0}}'
p1outputMode = '{{p1OutputMode#PROD-P1: output mode#Local}}'
p1eventsRequested = '{{p1EventsRequested#PROD-P1: events requested (-1 = ALL)#-1}}'
p1ancestorDepth = int('{{p1AncestorDepth#PROD-P1: ancestor depth#0}}')
try:
  p1compressionLvl = '{{p1CompressionLevel#PROD-P1: Compression Level per step, e.g. LOW,HIGH]#LOW}}'
except SyntaxError:
  p1compressionLvl = []
p1OutputVisFlag = ast.literal_eval(
    '{{p1OutputVisFlag#PROD-P1: Visibility flag of output files (a dict {"n step":"flag"})#{"1":"N"} }}')
p1OutputVisFlag = fillVisList(p1OutputVisFlag, len(pr.stepsInProds[0]))
try:
  p1VisFlagSpecial = '{{p1OutputVisFlagSpecial'
  p1VisFlagSpecial += '#PROD-P1: Special Visibility flag of output files (a dict {"n step":{"FType":"flag" } })#}}'
  p1OutputVisFlagSpecial = ast.literal_eval(p1VisFlagSpecial)
except SyntaxError:
  p1OutputVisFlagSpecial = {}

# p2 params
p2Plugin = '{{p2PluginType#PROD-P2: production plugin name#LHCbStandard}}'
p2Priority = int('{{p2Priority#PROD-P2: priority#2}}')
p2CPU = '{{p2MaxCPUTime#PROD-P2: Max CPU time in secs#1000000}}'
p2GroupSize = '{{p2GroupSize#PROD-P2: Group Size#1}}'
p2DataSE = '{{p2DataSE#PROD-P2: Output Data Storage Element#Tier1-DST}}'
try:
  p2DataSESpecial = ast.literal_eval('{{p2DataSESpecial#PROD-P2: Special SEs per file type, e.g. {"T1":"SE1"}#}}')
except SyntaxError:
  p2DataSESpecial = {}
p2Policy = '{{p2Policy#PROD-P2: data policy (download or protocol)#download}}'
p2RemoveInputs = ast.literal_eval('{{p2RemoveInputs#PROD-P2: removeInputs flag#False}}')
p2StepMask = '{{p2StepMask#PROD-P2: step output to save, semicolon separated (default is last#}}'
p2FileMask = '{{p2FileMask#PROD-P2: file types in output to save, semicolon separated (default is all)#}}'
p2multicoreFlag = '{{p2MulticoreFLag#PROD-P2: multicore flag#True}}'
p2NumberOfProcessors = '{{p2NumberOfProcessors#PROD-P2: jobs min/max n of processors#0,0}}'
p2outputMode = '{{p2OutputMode#PROD-P2: output mode#Local}}'
p2eventsRequested = '{{p2EventsRequested#PROD-P2: events requested (-1 = ALL)#-1}}'
p2ancestorDepth = int('{{p2AncestorDepth#PROD-P2: ancestor depth#0}}')
try:
  p2compressionLvl = '{{p2CompressionLevel#PROD-P2: Compression Level per step, e.g. LOW,HIGH#LOW}}'
except SyntaxError:
  p2compressionLvl = []
p2OutputVisFlag = ast.literal_eval(
    '{{p2OutputVisFlag#PROD-P2: Visibility flag of output files (a dict {"n step":"flag"})#{"2":"Y"} }}')
p2OutputVisFlag = fillVisList(p2OutputVisFlag, len(pr.stepsInProds[1]))
try:
  p2VisFlagSpecial = '{{p2OutputVisFlagSpecial'
  p2VisFlagSpecial += '#PROD-P2: Special Visibility flag of output files (a dict {"n step":{"FType":"flag" } })#}}'
  p2OutputVisFlagSpecial = ast.literal_eval(p2VisFlagSpecial)
except SyntaxError:
  p2OutputVisFlagSpecial = {}

# p3 params
p3Plugin = '{{p3PluginType#PROD-P3: production plugin name#LHCbStandard}}'
p3Priority = int('{{p3Priority#PROD-P3: priority#2}}')
p3CPU = '{{p3MaxCPUTime#PROD-P3: Max CPU time in secs#1000000}}'
p3GroupSize = '{{p3GroupSize#PROD-P3: Group Size#1}}'
p3DataSE = '{{p3DataSE#PROD-P3: Output Data Storage Element#Tier1-DST}}'
try:
  p3DataSESpecial = ast.literal_eval('{{p3DataSESpecial#PROD-P3: Special SEs per file type, e.g. {"T1":"SE1"}#}}')
except SyntaxError:
  p3DataSESpecial = {}
p3Policy = '{{p3Policy#PROD-P3: data policy (download or protocol)#download}}'
p3RemoveInputs = ast.literal_eval('{{p3RemoveInputs#PROD-P3: removeInputs flag#False}}')
p3StepMask = '{{p3StepMask#PROD-P3: step output to save, semicolon separated (default is last#}}'
p3FileMask = '{{p3FileMask#PROD-P3: file types in output to save, semicolon separated (default is all)#}}'
p3multicoreFlag = '{{p3MulticoreFLag#PROD-P3: multicore flag#True}}'
p3NumberOfProcessors = '{{p3NumberOfProcessors#PROD-P3: jobs min/max n of processors#0,0}}'
p3outputMode = '{{p3OutputMode#PROD-P3: output mode#Any}}'
p3eventsRequested = '{{p3EventsRequested#PROD-P3: events requested (-1 = ALL)#-1}}'
p3ancestorDepth = int('{{p3AncestorDepth#PROD-P3: ancestor depth#0}}')
try:
  p3compressionLvl = '{{p3CompressionLevel#PROD-P3: Compression Level per step, e.g. ["LOW","HIGH"]#HIGH}}'
except SyntaxError:
  p3compressionLvl = []
p3OutputVisFlag = ast.literal_eval(
    '{{p3OutputVisFlag#PROD-P3: Visibility flag of output files (a dict {"n step":"flag"})#{"3":"Y"} }}')
p3OutputVisFlag = fillVisList(p3OutputVisFlag, len(pr.stepsInProds[2]))
try:
  p3VisFlagSpecial = '{{p3OutputVisFlagSpecial'
  p3VisFlagSpecial += '#PROD-P3: Special Visibility flag of output files (a dict {"n step":{"FType":"flag" } })#}}'
  p3OutputVisFlagSpecial = ast.literal_eval(p3VisFlagSpecial)
except SyntaxError:
  p3OutputVisFlagSpecial = {}

parentReq = '{{_parent}}'
if not parentReq:
  pr.requestID = '{{ID}}'
else:
  pr.requestID = parentReq

pr.prodGroup = '{{inProPass}}' + '/' + '{{pDsc}}'
# used in case of a test e.g. certification etc.
pr.configName = '{{configName}}'
pr.configVersion = '{{configVersion}}'
pr.outConfigName = pr.configName
# Other parameters from the request page
pr.dqFlag = '{{inDataQualityFlag}}'  # UNCHECKED
pr.dataTakingConditions = '{{simDesc}}'
pr.processingPass = '{{inProPass}}'
if p1[0] == 1 and pr.prodsTypeList[0].lower() not in ('mcsimulation', 'mcfastsimulation'):
  pr.bkFileType = '{{inFileType}}'
  pr.bkQueries = ['Full'] if not p1NoBKQuery else ['']
elif p1[0] == 1 and pr.prodsTypeList[0].lower() in ('mcsimulation', 'mcfastsimulation'):
  pr.bkQueries = ['']
  pr.prodGroup = '{{pDsc}}'
elif p1[0] != 1 and pr.prodsTypeList[0].lower() not in ('mcsimulation', 'mcfastsimulation'):
  pr.bkQueries = ['fromPreviousProd']
  if not pr.previousProdID:
    gLogger.error("Please specify an input production")
    DIRACexit(2)
pr.eventType = '{{eventType}}'

if modulesList:
  pr.modulesList = modulesList.replace(' ', '').split(',')

if localTestFlag:
  pr.testFlag = True
  pr.publishFlag = False
  pr.prodsToLaunch = [1]

if validationFlag:
  pr.outConfigName = 'validation'

if enablePopularityReport:
  pr.extraModulesList = ['FileUsage']

inputDataList = []
if not pr.publishFlag:
  testData = 'LFN:/lhcb/LHCb/Collision11/CHARMCOMPLETEEVENT.DST'
  testData += '/00012586/0000/00012586_00000706_1.charmcompleteevent.dst'
  inputDataList.append(testData)
  p1Policy = 'protocol'
  evtsPerJob = '5'
  bkScriptFlag = True
else:
  bkScriptFlag = False

# In case we want just to test, we publish in the certification/test part of the BKK
if currentSetup == 'LHCb-Certification' or pr.testFlag:
  pr.outConfigName = 'certification'
  pr.configVersion = 'test'

if pr.testFlag:
  pr.startRun = 75336
  pr.endRun = 75340
  p1CPU = '100000'
  pr.events = ['2000']
  pr.dataTakingConditions = 'Beam3500GeV-VeloClosed-MagDown'
  pr.processingPass = 'Real Data/Reco12/Stripping17'
  pr.bkFileType = 'CHARMCOMPLETEEVENT.DST'
  pr.eventType = '90000000'
  pr.dqFlag = 'ALL'

if not p1StepMask:
  p1StepMask = len(pr.stepsInProds[0])
if not p2StepMask:
  try:
    p2StepMask = len(pr.stepsInProds[1])
  except IndexError:
    p2StepMask = ''
if not p3StepMask:
  try:
    p3StepMask = len(pr.stepsInProds[2])
  except IndexError:
    p3StepMask = ''

p1NumberOfProcessors = tuple(int(x) for x in p1NumberOfProcessors.replace(' ', '').split(','))
p2NumberOfProcessors = tuple(int(x) for x in p2NumberOfProcessors.replace(' ', '').split(','))
p3NumberOfProcessors = tuple(int(x) for x in p3NumberOfProcessors.replace(' ', '').split(','))

pr.outputSEs = [x for x in [p1DataSE, p2DataSE, p3DataSE] if x != '']
pr.specialOutputSEs = [p1DataSESpecial, p2DataSESpecial, p3DataSESpecial]
pr.removeInputsFlags = [p1RemoveInputs, p2RemoveInputs, p3RemoveInputs][0:len(pr.prodsTypeList)]
pr.priorities = [p1Priority, p2Priority, p3Priority][0:len(pr.prodsTypeList)]
pr.cpus = [p1CPU, p2CPU, p3CPU][0:len(pr.prodsTypeList)]
pr.groupSizes = [p1GroupSize, p2GroupSize, p3GroupSize][0:len(pr.prodsTypeList)]
pr.plugins = [p1Plugin, p2Plugin, p3Plugin][0:len(pr.prodsTypeList)]
pr.inputs = [inputDataList, [], []][0:len(pr.prodsTypeList)]
pr.inputDataPolicies = [p1Policy, p2Policy, p3Policy][0:len(pr.prodsTypeList)]
pr.outputFileSteps = [p1StepMask, p2StepMask, p3StepMask]
pr.outputFileMasks = [p1FileMask, p2FileMask, p3FileMask]
pr.multicore = [p1multicoreFlag, p2multicoreFlag, p3multicoreFlag]
pr.processors = [p1NumberOfProcessors, p2NumberOfProcessors, p3NumberOfProcessors]
pr.outputModes = [p1outputMode, p2outputMode, p3outputMode]
pr.events = [p1eventsRequested, p2eventsRequested, p3eventsRequested]
pr.ancestorDepths = [p1ancestorDepth, p2ancestorDepth, p3ancestorDepth]
# pr.compressionLvl = [p1compressionLvl] * len( pr.stepsInProds[0] ) +\
#                    [p2compressionLvl] * len( pr.stepsInProds[1] ) +\
#                    [p3compressionLvl] * len( pr.stepsInProds[2] )
pr.compressionLvl = [lvl for lvl in p1compressionLvl.split(',')] + [lvl for lvl in p2compressionLvl.split(',')] +\
                    [lvl for lvl in p3compressionLvl.split(',')]
pr.outputVisFlag = [p1OutputVisFlag, p2OutputVisFlag, p3OutputVisFlag]
pr.specialOutputVisFlag = [p1OutputVisFlagSpecial, p2OutputVisFlagSpecial, p3OutputVisFlagSpecial]

pr.buildAndLaunchRequest()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unittest for: LHCbDIRAC.AccountingSystem.private.Plotters.PopularityPlotter.

PopularityPlotter.__bases__:
  DIRAC.AccountingSystem.private.Plotters.BaseReporter

We are assuming there is a solid test of __bases__, we are not testing them
here and assuming they work fine.

IMPORTANT: the test MUST be pylint compliant !
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access

import unittest
from decimal import Decimal

import mock


class PopularityPlotterTestCase(unittest.TestCase):
  """PopularityPlotterTestCase."""

  moduleTested = None
  classsTested = None

  def mockModuleTested(self, moduleTested):
    """Used to not redo the mocking done on the parent class ( if any )"""

    # Tries to get the mocks of the parent TestCases ( if any )
    for baseClass in self.__class__.__bases__:
      try:
        # pylint: disable=no-member
        moduleTested = baseClass.mockModuleTested(moduleTested)
      except TypeError:
        continue

    # And then makes its own mock
    class MockPopularity(object):
      # pylint: disable=missing-docstring,no-init
      definitionKeyFields = ('DataType', 'Activity', 'FileType', 'Production',
                             'ProcessingPass', 'Conditions', 'EventType', 'StorageElement')

    moduleTested.Popularity = mock.Mock(return_value=MockPopularity())

    return moduleTested

  def setUp(self):
    """Setup the test case."""

    import LHCbDIRAC.AccountingSystem.private.Plotters.PopularityPlotter as moduleTested

    self.moduleTested = self.mockModuleTested(moduleTested)
    self.classsTested = self.moduleTested.PopularityPlotter

  def tearDown(self):
    """Tear down the test case."""

    del self.moduleTested
    del self.classsTested

# ...............................................................................


class PopularityPlotterUnitTest(PopularityPlotterTestCase):
  """PopularityPlotterUnitTest.

  <constructor>
   - test_instantiate
  <class variables>
   - test_typeName
   - test_typeKeyFields
   - test_reportDataUsageName
   - test_reportNormalizedDataUsageName
  <methods>
   - test_reportDataUsage
   - test_reportNormalizedDataUsage
   - test_plotDataUsage
   - test_plotNormalizedDataUsage
  """

  def test_instantiate(self):
    """tests that we can instantiate one object of the tested class."""
    obj = self.classsTested(None, None)
    self.assertEqual('PopularityPlotter', obj.__class__.__name__)

  def test_typeName(self):
    """test the class variable "_typeName"."""
    obj = self.classsTested(None, None)
    self.assertEqual(obj._typeName, "Popularity")

  def test_typeKeyFields(self):
    """test the class variable "_typeKeyFields"."""
    obj = self.classsTested(None, None)
    self.assertEqual(obj._typeKeyFields, ['DataType', 'Activity', 'FileType',
                                          'Production', 'ProcessingPass',
                                          'Conditions', 'EventType', 'StorageElement'
                                          ])

  def test_reportDataUsageName(self):
    """test the class variable "_reportDataUsageName"."""
    obj = self.classsTested(None, None)
    self.assertEqual(obj._reportDataUsageName, "Data Usage")

  def test_reportNormalizedDataUsageName(self):
    """test the class variable "_reportNormalizedDataUsageName"."""
    obj = self.classsTested(None, None)
    self.assertEqual(obj._reportNormalizedDataUsageName, "Normalized Data Usage")

  def test_reportDataUsage(self):
    """test the method "_reportDataUsage"."""

    mockAccountingDB = mock.MagicMock()
    mockAccountingDB._getConnection.return_value = {'OK': False, 'Message': 'No connection'}
    mockAccountingDB.retrieveBucketedData.return_value = {'OK': True, 'Value': []}
    mockAccountingDB.calculateBucketLengthForTime.return_value = 'BucketLength'
    obj = self.classsTested(mockAccountingDB, None)

    res = obj._reportDataUsage({'groupingFields': [0, ['mehh'], 'blah'],
                                'startTime': 'startTime',
                                'endTime': 'endTime',
                                'condDict': {}
                                })
#     self.assertEqual( res[ 'OK' ], False )
#     self.assertEqual( res[ 'Message' ], 'No connection' )

    # Changed mocked to run over different lines of code
    mockAccountingDB._getConnection.return_value = {'OK': True, 'Value': []}
    res = obj._reportDataUsage({'groupingFields': [0, ['mehh'], 'blah'],
                                'startTime': 'startTime',
                                'endTime': 'endTime',
                                'condDict': {}
                                })
    self.assertEqual(res['OK'], True)
    self.assertEqual(res['Value'], {'graphDataDict': {},
                                    'data': {},
                                    'unit': 'files',
                                    'granularity': 'BucketLength'
                                    })

    mockedData = (('90000000', 1355616000, 86400, Decimal('123456.789')),
                  ('90000000', 1355702400, 86400, Decimal('78901.2345')))
    mockAccountingDB.retrieveBucketedData.return_value = {'OK': True, 'Value': mockedData}
    mockAccountingDB.calculateBucketLengthForTime.return_value = 86400

    res = obj._reportDataUsage({'groupingFields': ('%s', ['EventType']),
                                'startTime': 1355663249.0,
                                'endTime': 1355749690.0,
                                'condDict': {'EventType': '90000000'}
                                })
    self.assertEqual(res['OK'], True)
    self.assertEqual(res['Value'], {'graphDataDict': {'90000000': {1355616000: 123.456789,
                                                                   1355702400: 78.901234500000001}
                                                      },
                                    'data': {'90000000': {1355616000: 123456.789,
                                                          1355702400: 78901.234500000006}
                                             },
                                    'unit': 'kfiles',
                                    'granularity': 86400
                                    })

  def test_reportNormalizedDataUsage(self):
    """test the method "_reportNormalizedDataUsage"."""

    mockAccountingDB = mock.Mock()
    mockAccountingDB._getConnection.return_value = {'OK': False, 'Message': 'No connection'}
    mockAccountingDB.retrieveBucketedData.return_value = {'OK': True, 'Value': []}
    mockAccountingDB.calculateBucketLengthForTime.return_value = 'BucketLength'
    obj = self.classsTested(mockAccountingDB, None)

#     res = obj._reportNormalizedDataUsage( { 'groupingFields' : [ 0, [ 'mehh' ], 'blah' ],
#                                             'startTime'      : 'startTime',
#                                             'endTime'        : 'endTime',
#                                             'condDict'       : {}
#                                            } )
#     self.assertEqual( res[ 'OK' ], False )
#     self.assertEqual( res[ 'Message' ], 'No connection' )

    # Changed mocked to run over different lines of code
    mockAccountingDB._getConnection.return_value = {'OK': True, 'Value': []}
    res = obj._reportNormalizedDataUsage({'groupingFields': [0, ['mehh'], 'blah'],
                                          'startTime': 'startTime',
                                          'endTime': 'endTime',
                                          'condDict': {}
                                          })
    self.assertEqual(res['OK'], True)
    self.assertEqual(res['Value'], {'graphDataDict': {},
                                    'data': {},
                                    'unit': 'files',
                                    'granularity': 'BucketLength'
                                    })

    mockedData = (('90000000', 1355616000, 86400, Decimal('123456.789')),
                  ('90000000', 1355702400, 86400, Decimal('78901.2345')),
                  ('90000001', 1355616000, 86400, Decimal('223456.789')),
                  ('90000001', 1355702400, 86400, Decimal('148901.2345')))
    mockAccountingDB.retrieveBucketedData.return_value = {'OK': True, 'Value': mockedData}
    mockAccountingDB.calculateBucketLengthForTime.return_value = 86400

    res = obj._reportNormalizedDataUsage({'groupingFields': ('%s', ['EventType']),
                                          'startTime': 1355663249.0,
                                          'endTime': 1355749690.0,
                                          'condDict': {'StorageElement': 'CERN'}
                                          })
    self.assertEqual(res['OK'], True)
    self.assertEqual(res['Value'], {'graphDataDict': {'90000001': {1355616000: 223.45678899999999,
                                                                   1355702400: 148.90123449999999},
                                                      '90000000': {1355616000: 123.456789,
                                                                   1355702400: 78.901234500000001}
                                                      },
                                    'data': {'90000001': {1355616000: 223456.78899999999,
                                                          1355702400: 148901.23449999999},
                                             '90000000': {1355616000: 123456.789,
                                                          1355702400: 78901.234500000006}
                                             },
                                    'unit': 'kfiles',
                                    'granularity': 86400
                                    })

# ...............................................................................


class PopularityPlotterUnitTestCrashes(PopularityPlotterTestCase):
  """PopularityPlotterUnitTestCrashes.

  <constructor>
   - test_instantiate
  <class variables>
  <methods>
  - test_reportDataUsage
  - test_reportNormalizedDataUsage
  - test_plotDataUsage
  - test_plotNormalizedDataUsage
  """

  def test_instantiate(self):
    """test the constructor."""

    self.assertRaises(TypeError, self.classsTested)
    self.assertRaises(TypeError, self.classsTested, None)
    self.assertRaises(TypeError, self.classsTested, None, None, None, None)

    self.assertRaises(TypeError, self.classsTested, extraArgs=None)
    self.assertRaises(TypeError, self.classsTested, None, extraArgs=None)
    self.assertRaises(TypeError, self.classsTested, None, None, None, extraArgs=None)

  def test_reportDataUsage(self):
    """test the method "_reportDataUsage"."""

    mockAccountingDB = mock.Mock()
    mockAccountingDB._getConnection.return_value = {'OK': False, 'Message': 'No connection'}
    mockAccountingDB.retrieveBucketedData.return_value = {'OK': True, 'Value': []}
    mockAccountingDB.calculateBucketLengthForTime.return_value = 'BucketLength'
    obj = self.classsTested(mockAccountingDB, None)

    self.assertRaises(KeyError, obj._reportDataUsage, {})
    self.assertRaises(IndexError, obj._reportDataUsage, {'groupingFields': []})
    self.assertRaises(TypeError, obj._reportDataUsage, {'groupingFields': [1, 2]})
    self.assertRaises(TypeError, obj._reportDataUsage, {'groupingFields': [1, [2]]})
    self.assertRaises(TypeError, obj._reportDataUsage, {'groupingFields': ['1', '2']})
    self.assertRaises(KeyError, obj._reportDataUsage, {'groupingFields': ['1', [2]]})
    self.assertRaises(KeyError, obj._reportDataUsage, {'groupingFields': ['1', [2, 2]],
                                                       'startTime': None})
    self.assertRaises(KeyError, obj._reportDataUsage, {'groupingFields': ['1', [2, 2]],
                                                       'startTime': None,
                                                       'endTime': None})
    self.assertRaises(TypeError, obj._reportDataUsage, {'groupingFields': ['1', [2, 2]],
                                                        'startTime': None,
                                                        'endTime': None,
                                                        'condDict': None})

  def test_reportNormalizedDataUsage(self):
    """test the method "_reportNormalizedDataUsage"."""

    mockAccountingDB = mock.Mock()
    mockAccountingDB._getConnection.return_value = {'OK': False, 'Message': 'No connection'}
    mockAccountingDB.retrieveBucketedData.return_value = {'OK': True, 'Value': []}
    mockAccountingDB.calculateBucketLengthForTime.return_value = 'BucketLength'
    obj = self.classsTested(mockAccountingDB, None)

    self.assertRaises(KeyError, obj._reportNormalizedDataUsage, {})
    self.assertRaises(IndexError, obj._reportNormalizedDataUsage, {'groupingFields': []})
    self.assertRaises(TypeError, obj._reportNormalizedDataUsage, {'groupingFields': [1, 2]})
    self.assertRaises(TypeError, obj._reportNormalizedDataUsage, {'groupingFields': [1, [2]]})
    self.assertRaises(TypeError, obj._reportNormalizedDataUsage, {'groupingFields': ['1', '2']})
    self.assertRaises(KeyError, obj._reportNormalizedDataUsage, {'groupingFields': ['1', [2]]})
    self.assertRaises(KeyError, obj._reportNormalizedDataUsage, {'groupingFields': ['1', [2, 2]],
                                                                 'startTime': None})
    self.assertRaises(KeyError, obj._reportNormalizedDataUsage, {'groupingFields': ['1', [2, 2]],
                                                                 'startTime': None,
                                                                 'endTime': None})
    self.assertRaises(TypeError, obj._reportNormalizedDataUsage, {'groupingFields': ['1', [2, 2]],
                                                                  'startTime': None,
                                                                  'endTime': None,
                                                                  'condDict': None})

  def test_plotDataUsage(self):
    """test the method "_plotDataUsage"."""

    obj = self.classsTested(None, None)
    self.assertRaises(TypeError, obj._plotDataUsage, None, None, None)
    self.assertRaises(KeyError, obj._plotDataUsage, {}, None, None)
    self.assertRaises(KeyError, obj._plotDataUsage, {'startTime': 'startTime'},
                      None, None)
    self.assertRaises(TypeError, obj._plotDataUsage, {'startTime': 'startTime',
                                                      'endTime': 'endTime'},
                      None, None)
    self.assertRaises(KeyError, obj._plotDataUsage, {'startTime': 'startTime',
                                                     'endTime': 'endTime'},
                      {}, None)
    self.assertRaises(KeyError, obj._plotDataUsage, {'startTime': 'startTime',
                                                     'endTime': 'endTime'},
                      {'granularity': 'granularity'}, None)
    self.assertRaises(KeyError, obj._plotDataUsage, {'startTime': 'startTime',
                                                     'endTime': 'endTime',
                                                     'grouping': 'grouping'},
                      {'granularity': 'granularity'}, None)
    self.assertRaises(KeyError, obj._plotDataUsage, {'startTime': 'startTime',
                                                     'endTime': 'endTime',
                                                     'grouping': 'grouping'},
                      {'granularity': 'granularity',
                       'graphDataDict': 'graphDataDict'}, None)

  def test_plotNormalizedDataUsage(self):
    """test the method "_plotNormalizedDataUsage"."""

    obj = self.classsTested(None, None)
    self.assertRaises(TypeError, obj._plotNormalizedDataUsage, None, None, None)
    self.assertRaises(KeyError, obj._plotNormalizedDataUsage, {}, None, None)
    self.assertRaises(KeyError, obj._plotNormalizedDataUsage, {'startTime': 'startTime'},
                      None, None)
    self.assertRaises(TypeError, obj._plotNormalizedDataUsage, {'startTime': 'startTime',
                                                                'endTime': 'endTime'},
                      None, None)
    self.assertRaises(KeyError, obj._plotNormalizedDataUsage, {'startTime': 'startTime',
                                                               'endTime': 'endTime'},
                      {}, None)
    self.assertRaises(KeyError, obj._plotNormalizedDataUsage, {'startTime': 'startTime',
                                                               'endTime': 'endTime'},
                      {'granularity': 'granularity'}, None)
    self.assertRaises(KeyError, obj._plotNormalizedDataUsage, {'startTime': 'startTime',
                                                               'endTime': 'endTime',
                                                               'grouping': 'grouping'},
                      {'granularity': 'granularity'}, None)
    self.assertRaises(KeyError, obj._plotNormalizedDataUsage, {'startTime': 'startTime',
                                                               'endTime': 'endTime',
                                                               'grouping': 'grouping'},
                      {'granularity': 'granularity',
                       'graphDataDict': 'graphDataDict'}, None)

################################################################################


if __name__ == '__main__':
  suite = unittest.defaultTestLoader.loadTestsFromTestCase(PopularityPlotterTestCase)
  suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(PopularityPlotterUnitTest))
  suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(PopularityPlotterUnitTestCrashes))
  testResult = unittest.TextTestRunner(verbosity=2).run(suite)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""DataProcessingProgressAgent."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import datetime
import os
import time
import shutil

from DIRAC import S_OK, gConfig, gLogger
from DIRAC.Core.Utilities.File import mkDir
from DIRAC.Core.Base.AgentModule import AgentModule

from LHCbDIRAC.ProductionManagementSystem.Client.ProcessingProgress import ProcessingProgress, HTMLProgressTable
from LHCbDIRAC.BookkeepingSystem.Client.BKQuery import BKQuery

__RCSID__ = "$Id$"

AGENT_NAME = 'ProductionManagement/DataProcessingProgressAgent'


class DataProcessingProgressAgent(AgentModule):

  #############################################################################
  def initialize(self):
    """Sets default values."""

    self.pollingTime = self.am_getOption('PollingTime', 6 * 60 * 60)
    self.printResult = self.am_getOption('Verbose', False)
    if self.printResult:
      gLogger.setLevel('VERBOSE')
    self.workDirectory = self.am_getWorkDirectory()
    self.statCollector = ProcessingProgress(os.path.join(self.workDirectory, "dirac-production-stats.pkl"))
    self.uploadDirectory = self.am_getOption('UploadDirectory', None)

    # Get back the loop number
    self.cacheFile = os.path.join(self.workDirectory, "cacheFile")
    try:
      cfile = open(self.cacheFile, 'r')
      self.iterationNumber = int(cfile.read())
      cfile.close()
    except BaseException:
      self.iterationNumber = 0

    # Get the list of processing passes
    reportList = self.am_getSection('ProgressReports')
    self.progressReports = {}
    for reportName in reportList:
      optionPath = os.path.join('ProgressReports', reportName)
      processingPasses = self.am_getOption(os.path.join(optionPath, 'ProcessingPass'), [])
      conditions = self.am_getOption(os.path.join(optionPath, 'ConditionDescription'), [])
      bkConfig = self.am_getOption(os.path.join(optionPath, 'BKConfig'), '/LHCb/Collision11')
      eventType = self.am_getOption(os.path.join(optionPath, 'EventType'), 90000000)
      fileType = self.am_getOption(os.path.join(optionPath, 'FileType'), 'BHADRON.DST')
      report = {'ConditionDescription': conditions}
      report['Frequency'] = self.am_getOption(os.path.join(optionPath, 'Frequency'), 1)
      report['HTMLFile'] = self.am_getOption(
          os.path.join(
              optionPath, 'HTMLFile'), reportName.replace(
              '.', '') + '-Progress' + os.path.extsep + 'htm')
      report['BKQuery'] = []
      report['ClearCache'] = self.am_getOption(os.path.join(optionPath, 'ClearCache'), [])
      for processingPass in processingPasses:
        bkPath = os.path.join(bkConfig, '*/Real Data', processingPass, str(eventType), fileType)
        bkQuery = BKQuery(bkPath, visible=False)
        if not bkQuery:
          self.log.error("Cannot build bkQuery for %s" % bkPath)
        else:
          report['BKQuery'].append(bkQuery)
      self.progressReports[reportName.replace('.', '/')] = report

    self.log.info("List of progress reports:")
    self.previousProdStats = {}
    for reportName in self.progressReports:
      printStr = '%s: ' % reportName
      for key, value in self.progressReports[reportName].items():
        if key != 'BKQuery':
          printStr += ", %s : %s" % (key, str(value))
        else:
          printStr += ", %s :" % key
          for bkQuery in self.progressReports[reportName]['BKQuery']:
            printStr += " - %s" % bkQuery.getPath()
      self.log.info(printStr)
    return S_OK()

  def execute(self):
    self.log.info("Now getting progress of processing (iteration %d)..." % self.iterationNumber)

    for reportName in sorted(self.progressReports):
      htmlTable = HTMLProgressTable(reportName.replace('.', '/'))
      reportLen = len(reportName) + 4
      self.log.info("\n%s\n* %s *\n%s" % (reportLen * '*', reportName, reportLen * '*'))
      report = self.progressReports[reportName]
      # Skip all by each "frequency" loop
      if report['Frequency'] == 0 or (self.iterationNumber % report['Frequency']) != 0:
        self.log.info("Skipping this iteration for %s" % reportName)
        continue
      self.statCollector.setClearCache(report['ClearCache'])
      summaryProdStats = []
      printOutput = ''
      outputHTML = os.path.join(self.workDirectory, report['HTMLFile'])
      for cond in report['ConditionDescription']:
        prodStats = 4 * [None]
        for bkQuery in report['BKQuery']:
          bkPath = bkQuery.getPath()
          if not bkQuery.getQueryDict():
            continue
          self.log.info(
              "\n=========================\nBookkeeping query %s\n=========================" %
              bkPath.replace(
                  '*', cond))
          bkQuery.setConditions(cond)
          stats = self.statCollector.getFullStats(bkQuery, printResult=self.printResult)
          processingPass = bkQuery.getProcessingPass().split('/')
          for ind in range(len(prodStats)):
            if not prodStats[ind]:
              prodStats[ind] = stats[ind]
            else:
              prodStats[ind] += stats[ind]
        summaryProdStats.append(prodStats)
        if self.printResult:
          printOutput += self.statCollector.outputResults(cond, reportName.replace('.', '/'), prodStats)
        htmlTable.writeHTML(cond, prodStats)

      if self.printResult:
        lines = printOutput.split('\n')
        for line in lines:
          self.log.info(line)

      if len(summaryProdStats) > 1:
        htmlTable.writeHTMLSummary(summaryProdStats)
      if reportName not in self.previousProdStats:
        x = self.statCollector.getPreviousStats(reportName)
        if x:
          self.previousProdStats[reportName] = x
      if reportName in self.previousProdStats:
        htmlTable.writeHTMLDifference(summaryProdStats, self.previousProdStats[reportName])
      else:
        print(reportName, 'not in previous stats')
      self.previousProdStats[reportName] = {"Time": time.ctime(time.time()), "ProdStats": summaryProdStats}
      self.statCollector.setPreviousStats(reportName, self.previousProdStats[reportName])
      try:
        fOpen = open(outputHTML, 'w')
        fOpen.write("<head>\n<title>Progress of %s</title>\n</title>\n" % bkQuery.getProcessingPass())
        fOpen.write(str(htmlTable.getTable()))
        fOpen.close()
        print("Successfully wrote HTML file", outputHTML)
        self.uploadHTML(outputHTML)
      except BaseException:
        print("Failed to write HTML file", outputHTML)

    # Save the loop number
    self.iterationNumber += 1
    fOpen = open(self.cacheFile, 'w')
    fOpen.write(str(self.iterationNumber))
    fOpen.close()
    return S_OK()

  def uploadHTML(self, htmlFile):
    if not self.uploadDirectory:
      return
    try:
      uploadDirBase = os.path.join(self.uploadDirectory, 'Daily', str(datetime.datetime.today()).split()[0])
      mkDir(uploadDirBase)
      i = 0
      while True:
        uploadDir = os.path.join(uploadDirBase, str(i))
        mkDir(uploadDir)
        uploadedFile = os.path.join(uploadDir, os.path.basename(htmlFile))
        if not os.path.exists(uploadedFile):
          break
        i += 1
      shutil.copy(htmlFile, uploadedFile)
      remoteLink = os.path.join(self.uploadDirectory, os.path.basename(htmlFile))
      if os.path.exists(remoteLink):
        os.remove(remoteLink)
      os.symlink(uploadedFile, remoteLink)
      print(htmlFile, "copied to", uploadedFile, "and link set at", remoteLink)
    except BaseException:
      print("Failed to upload", htmlFile, "to", self.uploadDirectory)

  def am_getSection(self, section):
    res = gConfig.getSections("%s/%s" % (self.am_getModuleParam('section'), section))
    if res['OK']:
      return res['Value']
    else:
      return []

################################################################################



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test class for plugins."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# imports
import unittest
import importlib
from mock import MagicMock

from DIRAC import gLogger

# sut
from LHCbDIRAC.TransformationSystem.Agent.TransformationPlugin import TransformationPlugin


paramsBase = {'AgentType': 'Automatic',
              'DerivedProduction': '0',
              'FileMask': '',
              'GroupSize': 1,
              'InheritedFrom': 0,
              'JobType': 'MCSimulation',
              'MaxNumberOfTasks': 0,
              'OutputDirectories': "['/lhcb/MC/20', '/lhcb/debug/20']",
              'OutputLFNs': "{'LogTargetPath': ['/lhcb/9.tar'], 'LogFilePath': ['/lhcb/9']}",
              'Priority': '0',
              'SizeGroup': '1',
              'Status': 'Active',
              'TransformationID': 1080,
              'Type': 'MCSimulation',
              'outputDataFileMask': 'GAUSSHIST;ALLSTREAMS.DST'}

data = {'/this/is/at.1': ['SE1'],
        '/this/is/at.2': ['SE2'],
        '/this/is/also/at.2': ['SE2'],
        '/this/is/at.12': ['SE1', 'SE2'],
        '/this/is/also/at.12': ['SE1', 'SE2'],
        '/this/is/at_123': ['SE1', 'SE2', 'SE3'],
        '/this/is/at_23': ['SE2', 'SE3'],
        '/this/is/at_4': ['SE4']}


class PluginsTestCase(unittest.TestCase):
  """Base class for the Agents test cases."""

  def setUp(self):
    self.mockTC = MagicMock()
    self.mockDM = MagicMock()
    self.mockRM = MagicMock()
    self.mockBKK = MagicMock()
    self.mockCatalog = MagicMock()
    self.mockCatalog.getFileSize.return_value()
    self.tPlugin = importlib.import_module('LHCbDIRAC.TransformationSystem.Agent.TransformationPlugin')
    self.tPlugin.TransformationClient = self.mockTC
    self.tPlugin.DataManager = self.mockDM
    self.tPlugin.FileCatalog = self.mockCatalog
    self.tPlugin.rmClient = self.mockRM
    self.maxDiff = None
    se = importlib.import_module('DIRAC.TransformationSystem.Client.Utilities')
    mockSE = MagicMock()
    se.StorageElement = mockSE

    gLogger.setLevel('DEBUG')

  def tearDown(self):
    #     sys.modules.pop( 'DIRAC.Core.Base.AgentModule' )
    #     sys.modules.pop( 'DIRAC.TransformationSystem.Agent.TransformationAgent' )
    pass


class PluginsBaseSuccess(PluginsTestCase):

  def test__LHCbStandard(self):
    # no input data, active
    pluginStandard = TransformationPlugin('LHCbStandard',
                                          transClient=self.mockTC,
                                          dataManager=self.mockDM,
                                          rmClient=self.mockRM,
                                          bkClient=self.mockBKK,
                                          fc=self.mockCatalog)
    pluginStandard.setParameters(paramsBase)
    res = pluginStandard.run()
    self.assertTrue(res['OK'])
    self.assertEqual(res['Value'], [])

    # input data, active
    pluginStandard = TransformationPlugin('LHCbStandard',
                                          transClient=self.mockTC,
                                          dataManager=self.mockDM,
                                          rmClient=self.mockRM,
                                          bkClient=self.mockBKK,
                                          fc=self.mockCatalog)
    pluginStandard.setParameters(paramsBase)
    pluginStandard.setInputData(data)
    params = {}
    params['TransformationID'] = 123
    params['Status'] = 'Active'
    params['GroupSize'] = 2
    pluginStandard.setParameters(params)
    res = pluginStandard.run()
    self.assertTrue(res['OK'])
    self.assertTrue(len(res['Value']) == 3)
    for t in res['Value']:
      self.assertTrue(len(t[1]) <= 2)

    # input data, flush
    pluginStandard = TransformationPlugin('LHCbStandard',
                                          transClient=self.mockTC,
                                          dataManager=self.mockDM,
                                          rmClient=self.mockRM,
                                          bkClient=self.mockBKK,
                                          fc=self.mockCatalog)
    pluginStandard.setParameters(paramsBase)
    pluginStandard.setInputData(data)
    params = {}
    params['TransformationID'] = 123
    params['Status'] = 'Flush'
    params['GroupSize'] = 2
    pluginStandard.setParameters(params)
    res = pluginStandard.run()
    self.assertTrue(res['OK'])
    self.assertTrue(len(res['Value']) == 6)
    for t in res['Value']:
      self.assertTrue(len(t[1]) <= 2)

#############################################################################
# Test Suite run
#############################################################################


if __name__ == '__main__':
  suite = unittest.defaultTestLoader.loadTestsFromTestCase(PluginsTestCase)
  suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(PluginsBaseSuccess))
  testResult = unittest.TextTestRunner(verbosity=2).run(suite)

# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""It interprets the XML reports and make a job, file, or replica object."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from xml.parsers.expat import ExpatError
from xml.dom.minidom import parse, parseString
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB import OracleBookkeepingDB
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.FileParam import FileParam
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.JobParameters import JobParameters
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.JobReader import JobReader
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.ReplicaReader import ReplicaReader
from LHCbDIRAC.BookkeepingSystem.DB.DataTakingConditionInterpreter import BeamEnergyCondition, \
    VeloCondition, \
    MagneticFieldCondition, \
    EcalCondition, \
    HcalCondition, \
    HltCondition, \
    ItCondition, \
    LoCondition, \
    MuonCondition, \
    OtCondition, \
    Rich1Condition, \
    Rich2Condition, \
    Spd_prsCondition, \
    TtCondition, \
    VeloPosition, \
    Context

__RCSID__ = "$Id$"


class XMLFilesReaderManager(object):
  """XMLFilesReaderManager class."""
  #############################################################################

  def __init__(self):
    """initialize the member of class."""
    self.jobReader_ = JobReader()
    self.replicaReader_ = ReplicaReader()

    self.bkClient_ = OracleBookkeepingDB()
    self.dm_ = DataManager()
    self.fileTypeCache = {}

    self.log = gLogger.getSubLogger('XMLFilesReaderManager')

  #############################################################################
  @staticmethod
  def readFile(filename):
    """reads an file content which format is XML."""
    try:
      with open(filename) as stream:
        doc = parse(stream)

      docType = doc.doctype  # job or replica
      xmltype = docType.name
    except NameError as ex:
      gLogger.error("XML reading error", filename)
      return S_ERROR(ex)

    return xmltype, doc, filename

  #############################################################################
  def readXMLfromString(self, xmlString):
    """read the xml string."""
    try:
      doc = parseString(xmlString)

      docType = doc.doctype  # job or replica
      xmltype = docType.name

      if xmltype == 'Replicas':
        replica = self.replicaReader_.readReplica(doc, "IN Memory")
        result = self.processReplicas(replica)
        del replica
        return result
      elif xmltype == 'Job':
        job = self.jobReader_.readJob(doc, "IN Memory")
        result = self.processJob(job)
        del job
        return result
      else:
        self.log.error("unknown XML file!!!")
    except ExpatError as ex:
      self.log.error("XML reading error", repr(ex))
      return S_ERROR(ex)

  #############################################################################
  def processJob(self, job):
    """interprets the xml content."""
    self.log.debug("Start Job Processing")

    # prepare for the insert, check the existence of the input files and retreive the fileid
    inputFiles = [inputFile.getFileName() for inputFile in job.getJobInputFiles()]
    if inputFiles:
      result = self.bkClient_.bulkgetIDsFromFilesTable(inputFiles)
      if not result['OK']:
        return result
      if result['Value']['Failed']:
        self.log.error("The following files are not in the bkk",
                       "%s" % (",".join(result['Value']['Failed'])))
        return S_ERROR("Files not in bkk")

      for inputFile in job.getJobInputFiles():
        lfn = inputFile.getFileName()
        fileID = int(result['Value']['Successful'][lfn]['FileId'])
        inputFile.setFileID(fileID)

    outputFiles = job.getJobOutputFiles()
    dqvalue = None
    for outputfile in outputFiles:

      typeName = outputfile.getFileType()
      typeVersion = outputfile.getFileVersion()
      cahedTypeNameVersion = typeName + '<<' + typeVersion
      if cahedTypeNameVersion in self.fileTypeCache:
        self.log.debug(cahedTypeNameVersion + ' in the cache!')
        typeID = self.fileTypeCache[cahedTypeNameVersion]
        outputfile.setTypeID(typeID)
      else:
        result = self.bkClient_.checkFileTypeAndVersion(typeName, typeVersion)
        if not result['OK']:
          self.log.error("The [type:version] is missing",
                         "[%s: %s]" % (str(typeName), str(typeVersion)))
          return S_ERROR("[type:version] missing")

        self.log.debug(cahedTypeNameVersion + " added to the cache!")
        typeID = int(result['Value'])
        outputfile.setTypeID(typeID)
        self.fileTypeCache[cahedTypeNameVersion] = typeID

      if job.getParam('JobType') and \
         job.getParam('JobType').getValue() == 'DQHISTOMERGING':  # all the merged histogram files have to be visible
        newFileParams = FileParam()
        newFileParams.setParamName('VisibilityFlag')
        newFileParams.setParamValue('Y')
        outputfile.addFileParam(newFileParams)
        self.log.debug('The Merged histograms visibility flag has to be Y!')

      params = outputfile.getFileParams()
      evtExists = False

      for param in params:
        paramName = param.getParamName()
        self.log.debug('ParamName check of ' + str(paramName))

        if paramName == "EventType":
          value = int(param.getParamValue())
          result = self.bkClient_.checkEventType(value)
          if not result['OK']:
            errorMessage = "The event type %s is missing!" % (str(value))
            return S_ERROR(errorMessage)

        if paramName == "EventTypeId":
          if param.getParamValue() != '':
            value = int(param.getParamValue())
            result = self.bkClient_.checkEventType(value)
            if not result['OK']:
              errorMessage = "The event type %s is missing!" % (str(value))
              return S_ERROR(errorMessage)
            evtExists = True

      if not evtExists and outputfile.getFileType() not in ['LOG']:
        inputFiles = job.getJobInputFiles()
        if inputFiles:
          fileName = inputFiles[0].getFileName()
          res = self.bkClient_.getFileMetadata([fileName])
          if res['OK']:
            fileMetadata = res['Value']['Successful'].get(fileName)
            if fileMetadata:
              if 'EventTypeId' in fileMetadata:
                if outputfile.exists('EventTypeId'):
                  param = outputfile.getParam('EventTypeId')
                  param.setParamValue(str(fileMetadata['EventTypeId']))
                else:
                  newFileParams = FileParam()
                  newFileParams.setParamName('EventTypeId')
                  newFileParams.setParamValue(str(fileMetadata['EventTypeId']))
                  outputfile.addFileParam(newFileParams)
            else:
              errMsg = "Can not get the metadata of %s file" % fileName
              self.log.error(errMsg)
              return S_ERROR(errMsg)
          else:
            return res
        elif job.getOutputFileParam('EventTypeId') is not None:
          param = job.getOutputFileParam('EventTypeId')
          newFileParams = FileParam()
          newFileParams.setParamName('EventTypeId')
          newFileParams.setParamValue(param.getParamValue())
          outputfile.addFileParam(newFileParams)
        else:
          return S_ERROR('It can not fill the EventTypeId because there is no input files!')

      infiles = job.getJobInputFiles()
      if not job.exists('RunNumber') and infiles:
        tck = -2
        runnumbers = []
        tcks = []
        for i in infiles:
          fileName = i.getFileName()
          retVal = self.bkClient_.getRunNbAndTck(fileName)

          if not retVal['OK']:
            return retVal
          if len(retVal['Value']) > 0:
            self.log.debug('RunTCK:', '%s' % retVal['Value'])

            for i in retVal['Value']:
              if i[0] not in runnumbers:
                runnumbers += [i[0]]
              if i[1] not in tcks:
                tcks += [i[1]]

          if len(runnumbers) > 1:
            self.log.warn('Different runs are reconstructed:', '%s' % runnumbers)
            runnumber = None
          else:
            runnumber = runnumbers[0]

          if len(tcks) > 1:
            self.log.warn('Different TCKs are reconstructed:', '%s' % tcks)
            tck = -2
          else:
            tck = tcks[0]

          self.log.debug('The output files of the job inherits the following run:', "%s" % runnumber)
          self.log.debug('The output files of the job inherits the following TCK:', '%s' % tck)

          if not job.exists('Tck'):
            newJobParams = JobParameters()
            newJobParams.setName('Tck')
            newJobParams.setValue(tck)
            job.addJobParams(newJobParams)

          if runnumber is not None:
            prod = None
            newJobParams = JobParameters()
            newJobParams.setName('RunNumber')
            newJobParams.setValue(str(runnumber))
            job.addJobParams(newJobParams)

            if job.getParam('JobType') and job.getParam('JobType').getValue() == 'DQHISTOMERGING':
              self.log.debug('DQ merging!')
              retVal = self.bkClient_.getJobInfo(fileName)
              if retVal['OK']:
                prod = retVal['Value'][0][18]
                newJobParams = JobParameters()
                newJobParams.setName('Production')
                newJobParams.setValue(str(prod))
                job.addJobParams(newJobParams)
                self.log.debug('Production inherited from input:', '%s' % prod)
            else:
              prod = job.getParam('Production').getValue()
              self.log.debug('Production:', '%s' % prod)

            retVal = self.bkClient_.getProductionProcessingPassID(prod)
            if retVal['OK']:
              proc = retVal['Value']

              retVal = self.bkClient_.getRunAndProcessingPassDataQuality(runnumber, proc)
              if retVal['OK']:
                dqvalue = retVal['Value']
              else:
                dqvalue = None
                message = (
                    "The rundataquality table does not contain run=%d proc_id=%s. Consequently, "
                    "the Dq flag is inherited from the ancestor file!"
                ) % (runnumber, proc)
                self.log.warn(message)
            else:
              dqvalue = None
              self.log.warn('Bkk can not set the quality flag because the processing \
              pass is missing for % d production (run number: %d )!' % (int(prod), int(runnumber)))

    inputfiles = job.getJobInputFiles()

    sumEventInputStat = 0
    sumEvtStat = 0
    sumLuminosity = 0

    if job.exists('JobType'):
      job.removeParam('JobType')

    # This must be replaced by a single call!!!!
    # ## It is not urgent as we do not have a huge load on the database
    for i in inputfiles:
      fname = i.getFileName()
      res = self.bkClient_.getJobInfo(fname)
      if not res['OK']:
        return res

      value = res["Value"]
      if value and value[0][2] is not None:
        sumEventInputStat += value[0][2]

      res = self.bkClient_.getFileMetadata([fname])
      if not res['OK']:
        return res

      fileMetadata = res['Value']['Successful'].get(fname)
      if fileMetadata:
        if fileMetadata['EventStat'] is not None:
          sumEvtStat += fileMetadata['EventStat']
        if fileMetadata['Luminosity'] is not None:
          sumLuminosity += fileMetadata['Luminosity']
        if dqvalue is None:
          dqvalue = fileMetadata.get('DataqualityFlag', fileMetadata.get('DQFlag', None))
      else:
        errMsg = "Can not get the metadata of %s file" % fname
        self.log.error(errMsg)
        return S_ERROR(errMsg)

    evtinput = 0
    if int(sumEvtStat) > int(sumEventInputStat):
      evtinput = sumEvtStat
    else:
      evtinput = sumEventInputStat

    if inputfiles:
      if not job.exists('EventInputStat'):
        newJobParams = JobParameters()
        newJobParams.setName('EventInputStat')
        newJobParams.setValue(str(evtinput))
        job.addJobParams(newJobParams)
      else:
        currentEventInputStat = job.getParam('EventInputStat')
        currentEventInputStat.setValue(evtinput)

    self.log.debug('Luminosity:', '%s' % sumLuminosity)
    outputFiles = job.getJobOutputFiles()
    for outputfile in outputFiles:
      if outputfile.getFileType() not in ['LOG'] and \
         sumLuminosity > 0 and not outputfile.exists('Luminosity'):
        newFileParams = FileParam()
        newFileParams.setParamName('Luminosity')
        newFileParams.setParamValue(sumLuminosity)
        outputfile.addFileParam(newFileParams)
        self.log.debug('Luminosity added to ', '%s' % outputfile.getFileName())
      ################

    config = job.getJobConfiguration()
    params = job.getJobParams()

    for param in params:
      if param.getName() == "RunNumber":
        value = int(param.getValue())
        if value <= 0 and len(job.getJobInputFiles()) == 0:
          # The files which inherits the runs can be entered to the database
          return S_ERROR('The run number not greater 0!')

    result = self.__insertJob(job)

    if not result['OK']:
      errorMessage = "Unable to create Job: %s , %s, %s .\n Error: %s" % (str(config.getConfigName()),
                                                                          str(config.getConfigVersion()),
                                                                          str(config.getDate()),
                                                                          str(result['Message']))
      return S_ERROR(errorMessage)
    else:
      jobID = int(result['Value'])
      job.setJobId(jobID)

    if job.exists('RunNumber'):
      try:
        runnumber = int(job.getParam('RunNumber').getValue())
      except ValueError:
        runnumber = -1
      if runnumber != -1:
        self.log.verbose("Registering the run status for ", "Run number %s,  JobId %s" %
                         (runnumber, job.getJobId()))
        result = self.bkClient_.insertRunStatus(runnumber, job.getJobId(), "N")
        if not result['OK']:
          errorMessage = ("Unable to register run status", runnumber + result['Message'])
          self.log.error(errorMessage[0], errorMessage[1])
          res = self.bkClient_.deleteJob(job.getJobId())
          if not res['OK']:
            self.log.warn("Unable to delete job", str(job.getJobId()) + res['Message'])
          return S_ERROR(errorMessage[0])

        # we may using HLT2 output to flag the runs as a consequence we may flagged the
        # runs before they registered to the bookkeeping.
        # we can flag a run using the newrunquality table
        retVal = self.bkClient_.getProductionProcessingPassID(-1 * int(runnumber))
        if retVal['OK']:
          retVal = self.bkClient_.getRunAndProcessingPassDataQuality(runnumber, retVal['Value'])
          if retVal['OK']:
            dqvalue = retVal['Value']
            self.log.verbose("The run data quality flag for", "run %d is %s" % (runnumber, dqvalue))
          else:
            # The report will be entered to the db.
            self.log.warn(retVal['Message'])
        else:
          self.log.error(retVal['Message'])
      else:
        # we reconstruct multiple runs
        self.log.warn("Run number can not determined for production:", job.getParam('Production').getValue())

    inputFiles = job.getJobInputFiles()
    for inputfile in inputFiles:
      result = self.bkClient_.insertInputFile(job.getJobId(), inputfile.getFileID())
      if not result['OK']:
        errorMessage = ("Unable to insert input file",
                        (str(inputfile.getFileName())) + result['Message'])
        self.log.error(errorMessage[0], errorMessage[1])
        res = self.bkClient_.deleteJob(job.getJobId())
        if not res['OK']:
          self.log.warn("Unable to delete job", str(job.getJobId()) + res['Message'])
        return S_ERROR(errorMessage[0])

    outputFiles = job.getJobOutputFiles()
    prod = job.getParam('Production').getValue()
    stepid = job.getParam('StepID').getValue()
    retVal = self.bkClient_.getProductionOutputFileTypes(prod, stepid)
    if not retVal['OK']:
      return retVal
    outputFileTypes = retVal['Value']
    for outputfile in outputFiles:
      if dqvalue is not None:
        newFileParams = FileParam()
        newFileParams.setParamName('QualityId')
        newFileParams.setParamValue(dqvalue)
        outputfile.addFileParam(newFileParams)
      if not job.exists('RunNumber'):  # if it is MC
        newFileParams = FileParam()
        newFileParams.setParamName('QualityId')
        newFileParams.setParamValue('OK')
        outputfile.addFileParam(newFileParams)
      ftype = outputfile.getFileType()
      if ftype in outputFileTypes:
        vFileParams = FileParam()
        vFileParams.setParamName('VisibilityFlag')
        vFileParams.setParamValue(outputFileTypes[ftype])
        outputfile.addFileParam(vFileParams)
        self.log.debug('The visibility flag is:' + outputFileTypes[ftype])

      result = self.__insertOutputFiles(job, outputfile)
      if not result['OK']:
        errorMessage = ("Unable to insert output file",
                        "%s ! ERROR: %s" % (str(outputfile.getFileName()),
                                            result["Message"]))
        self.log.error(errorMessage[0], errorMessage[1])
        res = self.bkClient_.deleteInputFiles(job.getJobId())
        if not res['OK']:
          self.log.warn("Unable to delete inputfiles of", str(job.getJobId()) + res['Message'])
        res = self.bkClient_.deleteJob(job.getJobId())
        if not res['OK']:
          self.log.warn("Unable to delete job", str(job.getJobId()) + res['Message'])
        return S_ERROR(errorMessage[0])
      else:
        fileid = int(result['Value'])
        outputfile.setFileID(fileid)

      replicas = outputfile.getReplicas()
      for replica in replicas:
        params = replica.getaprams()
        for param in params:  # just one param exist in params list, because JobReader only one param add to Replica
          name = param.getName()
        result = self.bkClient_.updateReplicaRow(outputfile.getFileID(), 'No')  # , name, location)
        if not result['OK']:
          errorMessage = "Unable to create Replica %s !" % (str(name))
          return S_ERROR(errorMessage)

    self.log.debug("End Processing!")

    return S_OK()

  def __insertJob(self, job):
    """Inserts the job to the database."""
    config = job.getJobConfiguration()

    production = None

    condParams = job.getDataTakingCond()
    if condParams is not None:
      datataking = condParams.getParams()
      config = job.getJobConfiguration()

      ver = config.getConfigVersion()  # online bug fix
      ver = ver.capitalize()
      config.setConfigVersion(ver)
      self.log.debug("Data taking:", "%s" % datataking)
      context = Context(datataking, config.getConfigName())
      conditions = [BeamEnergyCondition(), VeloCondition(),
                    MagneticFieldCondition(), EcalCondition(),
                    HcalCondition(), HltCondition(),
                    ItCondition(), LoCondition(),
                    MuonCondition(), OtCondition(),
                    Rich1Condition(), Rich2Condition(),
                    Spd_prsCondition(), TtCondition(),
                    VeloPosition()]
      for condition in conditions:
        condition.interpret(context)

      self.log.debug(context.getOutput())
      datataking['Description'] = context.getOutput()

      res = self.bkClient_.getDataTakingCondDesc(datataking)
      dataTackingPeriodDesc = None
      if res['OK']:
        daqid = res['Value']
        if len(daqid) != 0:  # exist in the database datataking
          dataTackingPeriodDesc = res['Value'][0][0]
          self.log.debug('Data taking condition id', dataTackingPeriodDesc)
        else:
          res = self.bkClient_.insertDataTakingCond(datataking)
          if not res['OK']:
            return S_ERROR("DATA TAKING Problem:" + str(res['Message']))
          dataTackingPeriodDesc = datataking['Description']
          # The new data taking condition inserted. The name should be the generated name.
      else:
        # Note we allow to insert data quality tags when only the description is different.
        res = self.bkClient_.insertDataTakingCond(datataking)
        if not res['OK']:
          return S_ERROR("DATA TAKING Problem:" + str(res['Message']))
        dataTackingPeriodDesc = datataking['Description']
        # The new data taking condition inserted. The name should be the generated name.

      # insert processing pass
      programName = None
      programVersion = None
      conddb = None
      dddb = None
      found = False
      for param in job.getJobParams():
        if param.getName() == 'ProgramName':
          programName = param.getValue()
        elif param.getName() == 'ProgramVersion':
          programVersion = param.getValue()
        elif param.getName() == 'CondDB':
          conddb = param.getValue()
        elif param.getName() == 'DDDB':
          dddb = param.getValue()
        elif param.getName() == 'RunNumber':
          production = int(param.getValue()) * -1
          found = True

      if job.exists('CondDB'):
        job.removeParam('CondDB')
      if job.exists('DDDB'):
        job.removeParam('DDDB')

      if not found:
        self.log.error('Run number is missing!')
        return S_ERROR('Run number is missing!')

      retVal = self.bkClient_.getStepIdandNameForRUN(programName, programVersion, conddb, dddb)

      if not retVal['OK']:
        return retVal

      stepid = retVal['Value'][0]

      # now we have to get the list of eventtypes
      eventtypes = []
      for outputFiles in job.getJobOutputFiles():
        for outPutfileParam in outputFiles.getFileParams():
          outputFileParamName = outPutfileParam.getParamName()
          if outputFileParamName == "EventTypeId":
            eventtypes.append(int(outPutfileParam.getParamValue()))

      steps = {'Steps':
               [{'StepId': stepid,
                 'StepName': retVal['Value'][1],
                 'ProcessingPass':retVal['Value'][1],
                 'Visible':'Y',
                 'OutputFileTypes':[{'FileType': 'RAW'}]}]}

      self.log.debug('Pass_indexid', "%s" % steps)
      self.log.debug('Data taking', "%s" % dataTackingPeriodDesc)
      self.log.debug('production', production)

      newJobParams = JobParameters()
      newJobParams.setName('StepID')
      newJobParams.setValue(str(stepid))
      job.addJobParams(newJobParams)

      message = "StepID for run: %s" % (str(production))
      self.log.info(message, stepid)

      res = self.bkClient_.addProduction(production,
                                         simcond=None,
                                         daq=dataTackingPeriodDesc,
                                         steps=steps['Steps'],
                                         inputproc='',
                                         configName=config.getConfigName(),
                                         configVersion=config.getConfigVersion(),
                                         eventType=eventtypes)
      if res['OK']:
        self.log.verbose("New processing pass has been created!")
        self.log.verbose("New production is:", production)
      elif job.exists('RunNumber'):
        self.log.warn('The run already registered!')
      else:
        self.log.error("Failing adding production", production + res['Message'])
        retVal = self.bkClient_.deleteStepContainer(production)
        if not retVal['OK']:
          return retVal
        return S_ERROR('Failing adding production')

    attrList = {'ConfigName': config.getConfigName(),
                'ConfigVersion': config.getConfigVersion(),
                'JobStart': None}

    for param in job.getJobParams():
      attrList[str(param.getName())] = param.getValue()

    res = self.bkClient_.checkProcessingPassAndSimCond(attrList['Production'])
    if not res['OK']:
      self.log.error('check processing pass and simulation condition error', res['Message'])
    else:
      value = res['Value']
      if value[0][0] == 0:
        errorMessage = "Missing processing pass and simulation conditions: "
        errorMessage += "please fill it. Production = %s" % (str(attrList['Production']))
        self.log.warn(errorMessage)

    if attrList['JobStart'] is None:
      # date = config.getDate().split('-')
      # time = config.getTime().split(':')
      # dateAndTime = datetime.datetime(int(date[0]), int(date[1]), int(date[2]), int(time[0]), int(time[1]), 0, 0)
      attrList['JobStart'] = config.getDate() + ' ' + config.getTime()

    if production is not None:  # for the online registration
      attrList['Production'] = production

    res = self.bkClient_.insertJob(attrList)

    if not res['OK'] and production < 0:
      self.log.error("Failed inserting job", res['Message'])
      retVal = self.bkClient_.deleteProductionsContainer(production)
      if not retVal['OK']:
        self.log.error(retVal['Message'])
    return res

  #############################################################################
  def __insertOutputFiles(self, job, outputfile):
    """insert the files produced by a job."""
    attrList = {'FileName': outputfile.getFileName(),
                'FileTypeId': outputfile.getTypeID(),
                'JobId': job.getJobId()}

    fileParams = outputfile.getFileParams()
    for param in fileParams:
      attrList[str(param.getParamName())] = param.getParamValue()
    return self.bkClient_.insertOutputFile(attrList)

  #############################################################################
  def processReplicas(self, replica):
    """process the replica registration request."""
    outputfile = replica.getFileName()
    self.log.debug("Processing replicas:", "%s" % outputfile)
    fileID = -1

    params = replica.getaprams()
    delete = True

    replicaFileName = ""
    for param in params:
      replicaFileName = param.getFile()
      location = param.getLocation()
      delete = param.getAction() == "Delete"

      result = self.bkClient_.checkfile(replicaFileName)
      if not result['OK']:
        message = "No replica can be "
        if delete:
          message += "removed"
        else:
          message += "added"
        message += " to file " + str(replicaFileName) + " for " + str(location) + ".\n"
        return S_ERROR(message)
      else:
        fileID = int(result['Value'][0][0])
        self.log.debug("FileId:", fileID)

      if delete:
        result = self.dm_.getReplicas(replicaFileName)
        replicaList = result['Value']['Successful']
        if len(replicaList) == 0:
          result = self.bkClient_.updateReplicaRow(fileID, "No")
          if not result['OK']:
            self.log.warn("Unable to set the Got_Replica flag for ", "%s" % replicaFileName)
            return S_ERROR("Unable to set the Got_Replica flag for ", "%s" % replicaFileName)
      else:
        result = self.bkClient_.updateReplicaRow(fileID, "Yes")
        if not result['OK']:
          return S_ERROR("Unable to set the Got_Replica flag for " + str(replicaFileName))

    self.log.debug("End Processing replicas!")

    return S_OK()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""ProdConf is a utility to manipulate a ProdConf file.

If the file does not exist, it will be created. If it exists and has
options, new ones will be put in if not existing, or override the old
ones if already existing. This is used by the production API to create
production workflows but also provides lists of options files for test
jobs.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import re


class ProdConf(object):
  """Class for managing ProdConf objects."""

  def __init__(self, fileName='prodConf.py', log=None):
    """initialize a ProdConf object, setting some relevant info."""

    self.optionsDict = {'Application': 'string',
                        'AppVersion': 'string',
                        'OptionFormat': 'string',
                        'InputFiles': 'list',
                        'OutputFilePrefix': 'string',
                        'OutputFileTypes': 'list',
                        'XMLFileCatalog': 'string',
                        'XMLSummaryFile': 'string',
                        'HistogramFile': 'string',
                        'DDDBTag': 'string',
                        'CondDBTag': 'string',
                        'DQTag': 'string',
                        'NOfEvents': 'integer',
                        'RunNumber': 'integer',
                        'FirstEventNumber': 'integer',
                        'TCK': 'string',
                        'ProcessingPass': 'string'}

    if not log:
      from DIRAC import gLogger
      self.log = gLogger.getSubLogger('ProdConf')
    else:
      self.log = log

    self.fileName = fileName

    if not os.path.exists(fileName):
      self.log.verbose('Creating ProdConf file %s from scratch' % fileName)
      with open(fileName, 'a'):
        os.utime(fileName, None)

    self.whatsIn = {}
    self._getWhatsIn()

  def _getWhatsIn(self):
    """Get what's in, as options, and fill the dictionary."""

    with open(self.fileName, 'r') as fopen:
      fileString = fopen.read()

    lines = re.split('\n+', fileString)
    for line in lines:
      for option, pcType in self.optionsDict.items():
        if re.match('[ ]*' + option + '[a-z,A-Z,0-9.]*', line):
          optionValues = re.split(option + '=+', line)
          for optionValue in optionValues:
            optionValue = optionValue.strip(' ')
            if optionValue:
              if pcType == 'list':
                optionValueEls = optionValue.split('[')
              else:
                optionValueEls = optionValue.split(',')
              for optionValueEl in optionValueEls:
                if optionValueEl:
                  value = optionValueEl.replace('"', '').replace(']', '').replace("'", '').strip(' ')
                  if pcType == 'list':
                    if value == ',':
                      value = []
                    else:
                      value = [x.strip() for x in value.split(',')]
                      value.remove('')
                  elif pcType == 'integer':
                    value = int(value)
                  self.whatsIn[option] = value

  def putOptionsIn(self, optionsDict, freshStart=False):
    """Put options, specified in the optionsDict, in the options file."""

    if freshStart:
      try:
        os.remove(self.fileName)
        self._getWhatsIn()
      except OSError:
        pass

    optsThatWillGoIn = self._buildOptions(optionsDict)
    stringToPut = self._getOptionsString(optsThatWillGoIn)
    self.log.debug("Going to write in %s" % self.fileName)
    self.log.debug(stringToPut)

    # Easier to re-write it completely
    with open(self.fileName, 'w') as fopen:
      fopen.write(stringToPut)

    self._getWhatsIn()

  def _buildOptions(self, optionsDict):
    """just build the options Dict."""
    optsThatWillGoIn = optionsDict
    for optAlreadyIn in self.whatsIn:
      if optAlreadyIn in optsThatWillGoIn:
        self.log.warn('Option %s of %s will be overwritten' % (optAlreadyIn, self.fileName))
      else:
        optsThatWillGoIn[optAlreadyIn] = self.whatsIn[optAlreadyIn]

    return optsThatWillGoIn

  def _getOptionsString(self, optsThatWillGoIn):
    """Build a string with the options that will go in."""
    string = 'from ProdConf import ProdConf\n\n'
    string = string + 'ProdConf(\n'
    for opt, value in optsThatWillGoIn.items():
      if self.optionsDict[opt] == 'list':
        string = string + '  ' + opt + '=' + str(value) + ',' + '\n'
      elif self.optionsDict[opt] == 'string':
        string = string + '  ' + opt + "='" + value + "'," + '\n'
      else:
        string = string + '  ' + opt + "=" + str(value) + "," + '\n'
    string = string + ')'

    return string



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Retrieve production summary from the Bookkeeping."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript, Script

  dmScript = DMScript()
  dmScript.registerBKSwitches()
  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s [option|cfgfile] ...' % Script.scriptName]))
  Script.parseCommandLine(ignoreErrors=True)

  from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
  bk = BookkeepingClient()

  bkQuery = dmScript.getBKQuery()
  if not bkQuery:
    gLogger.error("No BKQuery given...")
    DIRAC.exit(1)

  bkQueryDict = bkQuery.getQueryDict()
  dictItems = (
      'ConfigName',
      'ConfigVersion',
      'Production',
      'ConditionDescription',
      'ProcessingPass',
      'FileType',
      'EventType')
  for item in dictItems:
    bkQueryDict.setdefault(item, 'ALL')
  for item in list(bkQueryDict):
    if item not in dictItems:
      bkQueryDict.pop(item)

  gLogger.verbose('BKQuery:', bkQueryDict)
  res = bk.getProductionSummary(bkQueryDict)

  if not res["OK"]:
    gLogger.error(res["Message"])
    DIRAC.exit(1)

  records = res['Value']['Records']
  params = res['Value']['ParameterNames']
  width = 20

  gLogger.showHeaders(False)

  gLogger.notice('')
  gLogger.notice(
      params[0].ljust(30) + str(params[1]).ljust(30) +
      str(params[2]).ljust(30) + str(params[3]).ljust(30) +
      str(params[4]).ljust(30) + str(params[5]).ljust(30) +
      str(params[6]).ljust(20) + str(params[7]).ljust(20) +
      str(params[8]).ljust(20),
  )
  gLogger.notice('')
  for record in records:
    gLogger.notice(
        str(record[0]).ljust(15) + str(record[1]).ljust(15) +
        str(record[2]).ljust(20) + str(record[3]).ljust(width) +
        str(record[4]).ljust(width) + str(record[5]).ljust(width) +
        str(record[6]).ljust(width) + str(record[7]).ljust(width) +
        str(record[8]).ljust(width),
    )

  gLogger.notice('')
  gLogger.notice("TotalRecords = %d" % res['Value']['TotalRecords'])


if __name__ == "__main__":
  main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Start or stop the production(s)"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from DIRAC.Core.Base import Script

  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s <Production ID> |<Production ID>' % Script.scriptName,
      'Arguments:',
      '  <Production ID>:      DIRAC Production Id']))

  Script.registerSwitch('t', 'start', "Start the production")
  Script.registerSwitch('p', 'stop', "Stop the production")

  Script.parseCommandLine(ignoreErrors=True)

  args = Script.getPositionalArgs()
  if len(args) < 1:
    Script.showHelp(exitCode=2)

  from LHCbDIRAC.Interfaces.API.DiracProduction import DiracProduction

  Script.parseCommandLine(ignoreErrors=True)
  args = Script.getPositionalArgs()

  diracProd = DiracProduction()
  exitCode = 0
  errorList = []
  start = False
  stop = False
  action = ''

  switches = Script.getUnprocessedSwitches()

  for switch in switches:
    opt = switch[0].lower()

    if opt in ('t', 'start'):
      start = True
    if opt in ('p', 'stop'):
      stop = True

  if start and stop:
    print("ERROR: decide if you want to start or stop ( not both ).")
    DIRAC.exit(2)
  elif not (start or stop):
    print("ERROR: decide if you want to start or stop.")
    DIRAC.exit(2)
  elif start:
    action = 'start'
  elif stop:
    action = 'stop'

  for prodID in args:

    result = diracProd.production(prodID, action, disableCheck=False)
    if 'Message' in result:
      errorList.append((prodID, result['Message']))
      exitCode = 2
    elif not result:
      errorList.append((prodID, 'Null result for production() call'))
      exitCode = 2
    else:
      exitCode = 0

  for error in errorList:
    print("ERROR %s" % error)

  DIRAC.exit(exitCode)


if __name__ == "__main__":
  main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""check if a service or an agent is stalled."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
# TODO: This should be modernised to use subprocess(32)
try:
  import commands
except ImportError:
  # Python 3's subprocess module contains a compatibility layer
  import subprocess as commands

from DIRAC import gLogger
from DIRAC.Core.Utilities.DIRACScript import DIRACScript

msg = None


def write_log(mesg):
  """create the log file."""
  global msg
  gLogger.notice(mesg)
  msg += mesg + '\n'


@DIRACScript()
def main():
  global msg

  from DIRAC.Core.Base import Script
  Script.addDefaultOptionValue('LogLevel', 'verbose')
  Script.parseCommandLine(ignoreErrors=True)

  from DIRAC.Core.Utilities.Time import fromString, second, dateTime, timeInterval
  from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

  now = dateTime()
  runit_dir = '/opt/dirac/startup'
  logfile = 'log/current'
  pollingtime = 60
  diracAdmin = DiracAdmin()
  mailadress = 'lhcb-grid-shifter-oncall@cern.ch'
  subject = 'CheckStalled'
  host = os.uname()[1]
  msg = 'List of Services / agents which could be stalled on ' + host + ' \n\n'

  gLogger.getSubLogger("CheckStalledServices")
  write_log('The script ' + Script.scriptName + ' is running at ' + str(now))

  fd = os.listdir(runit_dir)
  # remove the web log as the format is different
  for dirname in fd:
    if dirname.find('Web') == 0:
      fd.remove(dirname)

  for dirname in fd:
    status, result = commands.getstatusoutput('runsvstat ' + runit_dir + '/' + dirname)
    filename = runit_dir + '/' + dirname + '/' + logfile
    with open(filename, "r") as fl:
      listLines = fl.readlines()
    lastLine = listLines[-1]
    write_log('Checking ----> ' + dirname)
    for line in listLines:
      if line.find("WARN: Server is not who"):
        if line.find("Polling time") != -1:
          try:
            pollingtime = line.split(':')[4].split(' ')[1].split('.')[0]
          except BaseException:
            try:
              pollingtime = line.split(':')[7].split(' ')[1].split('.')[0]
            except BaseException:
              write_log("    wrong format for Polling Time : " + line)
              break
        else:
          lastLine = line

    lastLineList = lastLine.split(' ')
    try:
      lastupdate = fromString(lastLineList[0] + ' ' + lastLineList[1])
    except BaseException:
      write_log('    EXCEPT : ' + dirname)
      write_log('   last line is ' + str(lastLineList))

    if isinstance(pollingtime, int):
      if int(pollingtime) < 59:
        pollingtime = 120

      interval = timeInterval(lastupdate, second * int(pollingtime))
      if not interval.includes(now):
        write_log("    the PollingTime is : " + str(pollingtime) + " s")
        write_log('    last update for ' + dirname + ' was : ' + str(lastupdate))
        write_log('    Polling Time is ' + str(pollingtime) + ' s')
        write_log('    last known status' + result + '\n')
        write_log('   Last line is : \n' + lastLine)

  res = diracAdmin.sendMail(mailadress, subject, msg, fromAddress='joel.closier@cern.ch')
  if not res['OK']:
    print('The mail could not be sent')


if __name__ == "__main__":
  main()
